package net.afyre.haha.demo.coq;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Writer;
import java.util.Hashtable;
import java.util.Scanner;

import net.afyre.haha.demo.coq.CoqExprPrinter.Priority;
import net.afyre.haha.demo.coq.CoqExprPrinter.PrioritySwitch;
import net.afyre.haha.demo.solver.context.Declaration;
import net.afyre.haha.demo.solver.context.FuncResultDeclaration;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.vcgen.DocumentedFormula;
import net.afyre.haha.demo.vcgen.VCContext;
import net.afyre.haha.demo.vcgen.VCProcessor;
import net.afyre.haha.demo.vcgen.VerificationCondition;

/**
 * Verification condition processor which prints
 * conditions and contexts in Coq format.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @author Aleksy Schubert <alx@mimuw.edu.pl>
 *
 */
public final class VCCoqPrinter extends VCProcessor {

  /** Output stream. */
  private final Writer output;

  /** Indentation controller */
  private final CoqIndent indent;

  /** Visitor used to print formulae. */
  private final CoqExprPrinter exprPrinter;

  /** Switch used to decide priority of formulae. */
  private PrioritySwitch prioritySwitch;

  /** Visitor used to print declarations. */
  private final CoqDeclarationPrinter declPrinter;
  
  /** The number of times a lemma name is used in the process. */
  private Hashtable<String, Integer> lemmaNames;

  /**
   * Constructor.
   * @param output Output stream.
   */
  public VCCoqPrinter(final Writer output) {
    this.output = output;
    this.indent = new CoqIndent(output);
    this.exprPrinter = new CoqExprPrinter(output);
    this.declPrinter = new CoqDeclarationPrinter(output, this.indent);
    this.prioritySwitch = new CoqExprPrinter.PrioritySwitch();
    printPrelude();
  }

  private void printPrelude() {
    print("Require Export ZArith ZArith.Int.\n");
    print("Local Open Scope Z_scope.\n");
    print("Load \"tactics.v\".\n\n");
    lemmaNames  = new Hashtable<String, Integer>();
  }

  // TODO: Move this to utils
  /**
   * Print an object (using toString()) to the output stream, 
   * converting any IOException to RuntimeExceptions.
   * 
   * @param obj Object to be printed (can be null).
   */
  private void print(final Object obj) {
    final String message;

    if (obj == null) {
      message = "null";
    } else {
      message = obj.toString();
    }
    try {
      output.write(message);
    } catch (final IOException e) {
      throw new RuntimeException(e);
    }
  }

  /** {@inheritDoc} */
  @Override
  protected void processCondition(final VerificationCondition vc) {
    indent.print();
    print("Lemma ");
    print(getLemmaName(vc));
    print(":\n");
    indent.inc();
    printConditionAssumptions(vc);
    print("\n");
    indent.print();
    print("->\n");
    indent.inc();
    indent.print();
    vc.getGoal().getFormula().accept(exprPrinter);
    print(".\n");
    indent.dec();
    indent.dec();
    indent.print();
    print("Proof.\n");
    indent.inc();
    indent.print();
    print("(* Give intros with meaningful names *)\n");
    indent.print();
    print("haha_solve.\n");
    indent.dec();
    indent.print();
    print("Admitted.\n");
    print("(** Change above to Qed.\n");
    print("    when the lemma is proved so that \n");
    print("    \"No more subgoals.\"\n");
    print("    occurs. *)\n");
    print("\n\n");
  }


  
  private String getLemmaName(final VerificationCondition vc) {
    final String name = vc.getGoal().getName();
    if (lemmaNames.containsKey(name)) {
      final int num = lemmaNames.get(name).intValue();
      lemmaNames.put(name, num + 1);
      return name + "_" + num;
    } else {
      lemmaNames.put(name, 1);
    }
    return name;
  }

  private boolean printConditionAssumptions(final VerificationCondition vc) {
    boolean first = true;
    for (final DocumentedFormula pre : vc.getAssumptions()) {
      if (!first) {
        print(" /\\\n");
      } else {
        first = false;
      }
      indent.print();
      final Formula formula = pre.getFormula();
      final int priority = formula.accept(prioritySwitch).getLeft();
      if (priority < Priority.AND) {
        print("("); //CoqOK
      }
      formula.accept(exprPrinter);
      if (priority < Priority.AND) {
        print(")"); //CoqOK
      }
    }
    return first;
  }

  /** {@inheritDoc} */
  @Override
  protected void pushContext(final VCContext context) {
    indent.print();
    print("(** Verification context: " + context.getComment() + "*)\n");
    indent.print();
    print("Module " + CoqLanguage.coqifyName(context.getName()) + ".\n\n");
    for (final Declaration decl : context.getDeclarations()) {
      indent.print();
      if (decl instanceof FuncResultDeclaration) {
        print("(* Variable with the function result. *)\n");
      }
      indent.print();
      decl.accept(declPrinter);
      print("\n");
    }
  }

  /** {@inheritDoc} */
  @Override
  protected void popContext(final VCContext context) {
    indent.print();
    print("End " + CoqLanguage.coqifyName(context.getName()) + ".\n");
  }

  /**
   * The method appends the text of Coq howto to the given Coq input.
   * 
   * @param inputStream the stream with the Coq howto text. 
   */
  public void addHowto(final InputStream inputStream) {
    final BufferedInputStream bs = new BufferedInputStream(inputStream);
    final Scanner scanner = new Scanner(bs);
    while (scanner.hasNextLine()) {
      final String line = scanner.nextLine();
      print(line);
      print("\n");
    }
    scanner.close();
    try {
      bs.close();
      inputStream.close();
    } catch (final IOException ex) {
      // We silently ignore the exception here.
    }
  }

}
