package net.afyre.haha.demo.solver.smt;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import net.afyre.haha.demo.solver.smt.sexpr.SExpr;
import net.afyre.haha.demo.solver.smt.sexpr.SExprPrinter;
import net.afyre.haha.demo.solver.smt.sexpr.SExprUtil;

/**
 * Class of models (counterexamples) produced by SMT solvers.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class SmtModel {
  /** Variable bindings. */
  private final Map<String, SExpr> bindings;
  
  /** Parameter lists for each variable. */
  private final Map<String, List<String>> parameters;
  
  /** Expression printer. */
  private final SExprPrinter printer;
 
  /** Constructor. */
  private SmtModel() {
    printer = new SExprPrinter();
    bindings = new TreeMap<String, SExpr>();
    parameters = new TreeMap<String, List<String>>();
  }
  
  /**
   * Create a model by parsing an expression returned
   * by Z3 after get-model call.
   * @param expr Expression returned by the solver.
   * @return Parsed model.
   * @throws IllegalArgumentException If given expression is not valid.
   * @throws NullPointerException If given expression is null.
   */
  public static SmtModel fromGetModel(final SExpr expr) {
    final SExprUtil util = SExprUtil.getInstance();
    
    if (expr == null) {
      throw new NullPointerException("expr");
    }
    
    final SmtModel result = new SmtModel();
    
    SExpr current = util.getTail(expr); // skip 'model'
    while (!util.isEmpty(current)) {
      SExpr it = util.getHead(current);
      it = util.getTail(it); // skip 'define-fun'
      final String name = util.toString(util.getHead(it));
      it = util.getTail(it);
      final List<String> arguments = parseArguments(util.getHead(it));
      // Skip type.
      it = util.getTail(util.getTail(it));
      final SExpr value = util.getHead(it);
      result.addBinding(name, arguments, value);
      
      current = util.getTail(current);
    }
    return result;
  }
  
  /**
   * Parse the argument list.
   * @param expr Argument list.
   * @return List of argument names.
   */
  private static List<String> parseArguments(final SExpr expr) {
    final SExprUtil util = SExprUtil.getInstance();
    final List<String> result = new LinkedList<String>();
    
    SExpr current = expr;
    while (!util.isEmpty(current)) {
      final SExpr binding = util.getHead(current);
      final String name = util.toString(util.getHead(binding));
      result.add(name);
      current = util.getTail(current);
    }
    
    return result;
  }
  
  /**
   * Add a new binding (used during model construction).
   * @param var Variable name.
   * @param arguments List of parameter names.
   * @param value Value (S-Expression).
   */
  private void addBinding(final String var, 
                          final List<String> arguments, 
                          final SExpr value) {
    bindings.put(var, value);
    parameters.put(var, new ArrayList<String>(arguments));
  }
  
  /**
   * Get all variables described by this model.
   * @return Immutable collection of variable names.
   */
  public Iterable<String> getVariables() {
    return Collections.unmodifiableSet(bindings.keySet());
  }
  
  /**
   * Get parameters of a variable.
   * @param var Variable name.
   * @return List of argument names.
   */
  public Iterable<String> getArguments(final String var) {
    return Collections.unmodifiableList(parameters.get(var));
  }
  
  /**
   * Get value of a variable.
   * @param var Variable.
   * @return Value (S-Expression) or null (if variable is not bound).
   */
  public SExpr getValue(final String var) {
    return bindings.get(var);
  }
  
  /** {@inheritDoc} */
  @Override
  public String toString() {
    final StringBuilder result = new StringBuilder();
    
    boolean first = true;
    for (final Entry<String, SExpr> entry : bindings.entrySet()) {
      if (!first) {
        result.append(", \n");
      }
      final String name = entry.getKey();
      result.append(name);
      final List<String> arguments = parameters.get(name);
      if (!arguments.isEmpty()) {
        result.append("(");
        final Iterator<String> it = arguments.iterator();
        result.append(it.next()); // list is not empty.
        while (it.hasNext()) {
          result.append(", ");
          result.append(it.next());
        }
        result.append(")");
      }
      result.append(" = ");
      result.append(entry.getValue().doSwitch(printer));
      first = false;
    }
    
    return result.toString();
  }
}
