package net.afyre.haha.demo.solver.smt.sexpr;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.regex.Pattern;

/**
 * Switch used to convert S-Expressions to strings.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class SExprPrinter implements SExprSwitch<String> {
  /** Singleton instance. */
  private static SExprPrinter instance;
  
  /** Pattern describing valid unquoted symbols. */
  private static final Pattern UNQUOTED_SYMBOL = 
      Pattern.compile("\\A[-\\p{Alpha}~!@$%^&*_+=<>.?/]" + 
                      "[-\\p{Alnum}~!@$%^&*_+=<>.?/]*\\z");
  
  /**
   * Create an instance of this class.
   * @return Printer object.
   */
  public static SExprPrinter getInstance() {
    if (instance == null) {
      instance = new SExprPrinter();
    }
    return instance;
  }
  
  /** {@inheritDoc} */
  @Override
  public String caseInteger(final BigInteger value) {
    return value.toString();
  }

  /** {@inheritDoc} */
  @Override
  public String caseDecimal(final BigDecimal value) {
    final String result = value.toPlainString();
    if (result.contains(".")) {
      return result;
    } else {
      return result + ".0";
    }
  }

  /** {@inheritDoc} */
  @Override
  public String caseHex(final BigInteger value, final int length) {
    if (length > 0) {
      return String.format("#x%0" + length + "x", value);
    } else {
      return "#x" + value.toString(16);
    }
  }

  /** {@inheritDoc} */
  @Override
  public String caseBin(final BigInteger value, final int length) {
    final StringBuilder result = new StringBuilder();
    
    result.append("#b");
    for (int i = 0; i < value.bitLength(); i++) {
      result.append('0');
    }
    result.append(value.toString(2));
    
    return result.toString();
  }

  /** {@inheritDoc} */
  @Override
  public String caseString(final String value) {
    final StringBuilder result = new StringBuilder(value.length());
    
    result.append('"');
    
    int i;
    final int n;
    i = 0;
    n = value.length();
    while (i < n) {
      final int ch = value.codePointAt(i);
      if (ch == '\\' || ch == '"') {
        result.append('\\');
      }
      result.appendCodePoint(ch);
      i = i + Character.charCount(ch);
    }
    
    result.append('"');
    
    return result.toString();
  }

  /** {@inheritDoc} */
  @Override
  public String caseSymbol(final String name) {
    if (UNQUOTED_SYMBOL.matcher(name).matches()) {
      return name;
    } else {
      return "|" + name + "|";
    }
  }

  /** {@inheritDoc} */
  @Override
  public String caseKeyword(final String name) {
    return ":" + name;
  }

  /** {@inheritDoc} */
  @Override
  public String caseList(final Iterable<SExpr> elements) {
    final StringBuilder result = new StringBuilder();
    
    result.append('(');
    int index = 0;
    for (final SExpr element : elements) {
      if (index > 0) {
        result.append(' ');
      }
      result.append(element.doSwitch(this));
      index++;
    }
    result.append(')');
    
    return result.toString();
  }
}
