package net.afyre.haha.demo.solver.smt.sexpr;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Iterator;

import net.afyre.haha.demo.util.OffsetIterable;

/**
 * Singleton class with utility functions
 * for manipulating S-Expressions.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class SExprUtil {
  /** Instance of this class. */
  private static final SExprUtil instance = new SExprUtil();
  
  /** Factory used to create new expressions. */
  private final SExprFactory factory;
  
  /** Switch used to print expressions. */
  private final SExprSwitch<String> printer;
  
  /** Switch used to extract head of a list. */
  private final SExprSwitch<SExpr> headSwitch;
  
  /** Switch used to extract tail of a list. */
  private final SExprSwitch<SExpr> tailSwitch;

  /** Switch used to check if a list is empty. */
  private final SExprSwitch<Boolean> isEmptySwitch;

  /** Constructor. */
  private SExprUtil() {
    factory = new DefaultSExprFactory();
    printer = new SExprPrinter();
    headSwitch = new HeadSwitch();
    tailSwitch = new TailSwitch();
    isEmptySwitch = new IsEmptySwitch();
  }

  
  /**
   * Create an instance of this class.
   * @return Instance.
   */
  public static SExprUtil getInstance() {
    return instance;
  }
  
  /**
   * Convert an expression to a string.
   * @param expr Expression.
   * @return Textual from of given expression.
   */
  public String toString(final SExpr expr) {
    if (expr == null) {
      return "null";
    } else {
      return expr.doSwitch(printer);
    }
  }
  
  /**
   * Get first element of a list expression.
   * @param expr Expression.
   * @return First element.
   * @throws IllegalArgumentException If the expression is not a list, 
   *                                  or is an empty list.
   * @throws NullPointerException  If expression is null.
   */
  public SExpr getHead(final SExpr expr) {
    if (expr == null) {
      throw new NullPointerException("expr");
    }
    return expr.doSwitch(headSwitch);
  }
  
  /**
   * Get tail of a list expression.
   * @param expr Expression.
   * @return Tail (new list).
   * @throws IllegalArgumentException If the expression is not a list, 
   *                                  or is an empty list.
   * @throws NullPointerException  If expression is null.
   */
  public SExpr getTail(final SExpr expr) {
    if (expr == null) {
      throw new NullPointerException("expr");
    }
    return expr.doSwitch(tailSwitch);
  }
  
  /**
   * Check if a list epxression represents the empty list.
   * @param expr Expression.
   * @return Tail (new list).
   * @throws IllegalArgumentException If the expression is not a list..
   * @throws NullPointerException  If expression is null.
   */
  public boolean isEmpty(final SExpr expr) {
    if (expr == null) {
      throw new NullPointerException("expr");
    }
    return expr.doSwitch(isEmptySwitch);
  }
  
  /** 
   * Abstract superclass of switches 
   * which implement list operations. 
   * 
   * @param <T> Type of values returned by the switch.
   **/
  private abstract class ListSwitch<T> implements SExprSwitch<T> {
    /**
     * Get operation name (used in error messages).
     * @return Function name.
     */
    protected abstract String getOpName();
    
    /** {@inheritDoc} */
    @Override
    public T caseInteger(final BigInteger value) {
      throw new IllegalArgumentException(
          getOpName() + ": list expected, got integer");
    }

    /** {@inheritDoc} */
    @Override
    public T caseDecimal(final BigDecimal value) {
      throw new IllegalArgumentException(
          getOpName() + ": list expected, got decinal");
    }

    /** {@inheritDoc} */
    @Override
    public T caseHex(final BigInteger value, final int length) {
      throw new IllegalArgumentException(
          getOpName() + ": list expected, got hex integer");
    }

    /** {@inheritDoc} */
    @Override
    public T caseBin(final BigInteger value, final int length) {
      throw new IllegalArgumentException(
          getOpName() + ": list expected, got binary integer");
    }

    /** {@inheritDoc} */
    @Override
    public T caseString(final String value) {
      throw new IllegalArgumentException(
          getOpName() + ": list expected, got string");    
    }

    /** {@inheritDoc} */
    @Override
    public T caseSymbol(final String name) {
      throw new IllegalArgumentException(
          getOpName() + ": list expected, got binary symbol");    
    }

    /** {@inheritDoc} */
    @Override
    public T caseKeyword(final String name) {
      throw new IllegalArgumentException(
          getOpName() + ": list expected, got keyword");    
    }
  }
  
  /** Switch used to extract head of a list. */
  private final class HeadSwitch extends ListSwitch<SExpr> {
    /** {@inheritDoc} */
    @Override
    protected String getOpName() {
      return "head";
    }
    
    /** {@inheritDoc} */
    @Override
    public SExpr caseList(final Iterable<SExpr> elements) {
      final Iterator<SExpr> it = elements.iterator();
      if (!it.hasNext()) {
        throw new IllegalArgumentException(
            getOpName() + ": empty list");
      }
      return it.next(); 
    } 
  }

  /** Switch used to extract tail of a list. */
  private final class TailSwitch extends ListSwitch<SExpr> {
    /** {@inheritDoc} */
    @Override
    protected String getOpName() {
      return "tail";
    }
    
    /** {@inheritDoc} */
    @Override
    public SExpr caseList(final Iterable<SExpr> elements) {
      final Iterator<SExpr> it = elements.iterator();
      if (!it.hasNext()) {
        throw new IllegalArgumentException(
            getOpName() + ": empty list");
      }
      return factory.createList(
                 OffsetIterable.getInstance(elements, 1)); 
    } 
  }
  
  /** Switch used to check if a list is empty. */
  private final class IsEmptySwitch extends ListSwitch<Boolean> {
    /** {@inheritDoc} */
    @Override
    protected String getOpName() {
      return "is_empty";
    }
    
    /** {@inheritDoc} */
    @Override
    public Boolean caseList(final Iterable<SExpr> elements) {
      final Iterator<SExpr> it = elements.iterator();
      return !it.hasNext(); 
    } 
  }
}
