package net.afyre.haha.demo.solver.expr.formulae;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * Interface of objects used to create formulae.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */

public interface FormulaFactory {
  /**
   * Create a formula of kind 'True'
   * @see FormulaSwitch#caseTrue
   * @return Expression object.
   */
  Formula createTrue();

  /**
   * Create a formula of kind 'False'
   * @see FormulaSwitch#caseFalse
   * @return Expression object.
   */
  Formula createFalse();

  /**
   * Create a formula of kind 'FormulaVariable'
   * @see FormulaSwitch#caseFormulaVariable
   * @param name {@link FormulaSwitch#caseFormulaVariable}
   * @return Expression object.
   */
  Formula createFormulaVariable(String name);

  /**
   * Create a formula of kind 'Negation'
   * @see FormulaSwitch#caseNegation
   * @param arg {@link FormulaSwitch#caseNegation}
   * @return Expression object.
   */
  Formula createNegation(Formula arg);

  /**
   * Create a formula of kind 'Conjunction'
   * @see FormulaSwitch#caseConjunction
   * @param left {@link FormulaSwitch#caseConjunction}
   * @param right {@link FormulaSwitch#caseConjunction}
   * @return Expression object.
   */
  Formula createConjunction(Formula left,
                            Formula right);

  /**
   * Create a formula of kind 'Disjunction'
   * @see FormulaSwitch#caseDisjunction
   * @param left {@link FormulaSwitch#caseDisjunction}
   * @param right {@link FormulaSwitch#caseDisjunction}
   * @return Expression object.
   */
  Formula createDisjunction(Formula left,
                            Formula right);

  /**
   * Create a formula of kind 'Implication'
   * @see FormulaSwitch#caseImplication
   * @param left {@link FormulaSwitch#caseImplication}
   * @param right {@link FormulaSwitch#caseImplication}
   * @return Expression object.
   */
  Formula createImplication(Formula left,
                            Formula right);

  /**
   * Create a formula of kind 'Xor'
   * @see FormulaSwitch#caseXor
   * @param left {@link FormulaSwitch#caseXor}
   * @param right {@link FormulaSwitch#caseXor}
   * @return Expression object.
   */
  Formula createXor(Formula left,
                    Formula right);

  /**
   * Create a formula of kind 'Equivalence'
   * @see FormulaSwitch#caseEquivalence
   * @param left {@link FormulaSwitch#caseEquivalence}
   * @param right {@link FormulaSwitch#caseEquivalence}
   * @return Expression object.
   */
  Formula createEquivalence(Formula left,
                            Formula right);

  /**
   * Create a formula of kind 'Eq'
   * @see FormulaSwitch#caseEq
   * @param left {@link FormulaSwitch#caseEq}
   * @param right {@link FormulaSwitch#caseEq}
   * @return Expression object.
   */
  Formula createEq(Term left,
                   Term right);

  /**
   * Create a formula of kind 'Neq'
   * @see FormulaSwitch#caseNeq
   * @param left {@link FormulaSwitch#caseNeq}
   * @param right {@link FormulaSwitch#caseNeq}
   * @return Expression object.
   */
  Formula createNeq(Term left,
                    Term right);

  /**
   * Create a formula of kind 'Distinct'
   * @see FormulaSwitch#caseDistinct
   * @param args {@link FormulaSwitch#caseDistinct}
   * @return Expression object.
   */
  Formula createDistinct(Term... args);

  /**
   * Create a formula of kind 'Lt'
   * @see FormulaSwitch#caseLt
   * @param left {@link FormulaSwitch#caseLt}
   * @param right {@link FormulaSwitch#caseLt}
   * @return Expression object.
   */
  Formula createLt(Term left,
                   Term right);

  /**
   * Create a formula of kind 'Gt'
   * @see FormulaSwitch#caseGt
   * @param left {@link FormulaSwitch#caseGt}
   * @param right {@link FormulaSwitch#caseGt}
   * @return Expression object.
   */
  Formula createGt(Term left,
                   Term right);

  /**
   * Create a formula of kind 'Le'
   * @see FormulaSwitch#caseLe
   * @param left {@link FormulaSwitch#caseLe}
   * @param right {@link FormulaSwitch#caseLe}
   * @return Expression object.
   */
  Formula createLe(Term left,
                   Term right);

  /**
   * Create a formula of kind 'Ge'
   * @see FormulaSwitch#caseGe
   * @param left {@link FormulaSwitch#caseGe}
   * @param right {@link FormulaSwitch#caseGe}
   * @return Expression object.
   */
  Formula createGe(Term left,
                   Term right);

  /**
   * Create a formula of kind 'Exists'
   * @see FormulaSwitch#caseExists
   * @param variables {@link FormulaSwitch#caseExists}
   * @param arg {@link FormulaSwitch#caseExists}
   * @param patterns {@link FormulaSwitch#caseExists}
   * @return Expression object.
   */
  Formula createExists(Iterable<VariableBinding> variables,
                       Formula arg,
                       Expression... patterns);

  /**
   * Create a formula of kind 'Forall'
   * @see FormulaSwitch#caseForall
   * @param variables {@link FormulaSwitch#caseForall}
   * @param arg {@link FormulaSwitch#caseForall}
   * @param patterns {@link FormulaSwitch#caseForall}
   * @return Expression object.
   */
  Formula createForall(Iterable<VariableBinding> variables,
                       Formula arg,
                       Expression... patterns);

  /**
   * Create a formula of kind 'Let'
   * @see FormulaSwitch#caseLet
   * @param var {@link FormulaSwitch#caseLet}
   * @param type {@link FormulaSwitch#caseLet}
   * @param value {@link FormulaSwitch#caseLet}
   * @param body {@link FormulaSwitch#caseLet}
   * @return Expression object.
   */
  Formula createFLet(String var,
                     ValueType type,
                     Term value,
                     Formula body);

  /**
   * Create a formula of kind 'LetProp'
   * @see FormulaSwitch#caseLetProp
   * @param var {@link FormulaSwitch#caseLetProp}
   * @param value {@link FormulaSwitch#caseLetProp}
   * @param body {@link FormulaSwitch#caseLetProp}
   * @return Expression object.
   */
  Formula createFLetProp(String var,
                         Formula value,
                         Formula body);

  /**
   * Create a formula of kind 'PredicateCall'
   * @see FormulaSwitch#casePredicateCall
   * @param pred {@link FormulaSwitch#casePredicateCall}
   * @param args {@link FormulaSwitch#casePredicateCall}
   * @return Expression object.
   */
  Formula createPredicateCall(String pred,
                              Term... args);
}
