package net.afyre.haha.demo.solver.expr.formulae;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * Interface of objects used to deconstruct formulae.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * 
 * @param <Result> Type of objects returned by visitor methods.
 */
public interface FormulaSwitch<Result> {
  /** 
   * True literal.
   * @return Switch result.
   */
  Result caseTrue();
  
  /** 
   * False literal.
   * @return Switch result. 
   */
  Result caseFalse();
  
  /** 
   * Variable.
   * @param name Variable name.
   * @return Switch result.
   */
  Result caseFormulaVariable(String name);
  
  /** 
   * Negation. 
   * @param arg Argument. 
   * @return Switch result.
   **/
  Result caseNegation(Formula arg);
  
  /**
   * Conjunction.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseConjunction(Formula left, Formula right);
  
  /**
   * Disjunction.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseDisjunction(Formula left, Formula right);
  
  /**
   * Implication.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseImplication(Formula left, Formula right);
  
  /**
   * Exclusive or.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseXor(Formula left, Formula right);
  
  /**
   * Equivalence.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseEquivalence(Formula left, Formula right); 
  
  /**
   * Equality test.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseEq(Term left, Term right); 
  
  /**
   * Negated equality test.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseNeq(Term left, Term right); 
  
  /**
   * Pairwise inequality test.
   * @param args Arguments.
   * @return Switch result.
   */
  Result caseDistinct(Term... args); 

  /**
   * Less than.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseLt(Term left, Term right); 

  /**
   * Greater than.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseGt(Term left, Term right); 

  /**
   * Less than or equal to.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseLe(Term left, Term right); 

  /**
   * Greater than or equal to.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result.
   */
  Result caseGe(Term left, Term right); 
  
  /**
   * Existential quantifier.
   * @param variables List of bound variables (must not be empty).
   * @param arg Formula.
   * @param patterns Instantiation patterns.
   * @return Switch result.
   */
  Result caseExists(Iterable<VariableBinding> variables, Formula arg, 
                    Expression... patterns);

  /**
   * Universal quantifier.
   * @param variables List of bound variables (must not be empty).
   * @param arg Formula.
   * @param patterns Instantiation patterns.
   * @return Switch result.
   */
  Result caseForall(Iterable<VariableBinding> variables, Formula arg, 
                    Expression... patterns);
  
  /**
   * Local variable binding.
   * @param var Variable name.
   * @param type Variable type.
   * @param value Assigned value.
   * @param body Expression body.
   * @return Switch result.
   */
  Result caseFLet(String var, ValueType type, 
                Term value, Formula body);
  
  /**
   * Local propositional variable binding.
   * @param var Variable name.
   * @param value Assigned value.
   * @param body Expression body.
   * @return Switch result.
   */
  Result caseFLetProp(String var, Formula value, Formula body);
  
  /**
   * Predicate call.
   * @param pred Predicate name.
   * @param args Arguments.
   * @return Switch result. 
   **/
  Result casePredicateCall(String pred, Term... args);
  
}
