package net.afyre.haha.demo.solver.expr.terms;

import java.math.BigInteger;

import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * Interface of objects used to deconstruct terms.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @param <Result> Type of objects returned by visitor methods.
 */
public interface TermSwitch<Result> {
  /** 
   * Integer literal. 
   * @param value Value of this literal. 
   * @return Switch result. 
   **/
  Result caseIntConst(BigInteger value);
  
  /**
   * Visit variable.
   * @param var Variable name.
   * @return Switch result. 
   **/
  Result caseVariable(String var);
  
  /**
   * Tuple construction.
   * @param elements Tuple elements.
   * @return Switch result. 
   **/
  Result caseMkTuple(Term... elements);
  
  /**
   * Tuple element access.
   * @param tuple Source tuple.
   * @param index Element index.
   * @return Switch result. 
   **/
  Result caseTupleAccess(Term tuple, int index);
  
  /**
   * Unary minus.
   * @param arg Argument.
   * @return Switch result. 
   **/
  Result caseUminus(Term arg);
  
  /**
   * Addition.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result. 
   **/
  Result caseAddition(Term left, Term right);
  
  /**
   * Subtraction.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result. 
   **/
  Result caseSubtraction(Term left, Term right);
  
  /**
   * Multiplication.
   * @param left Left argument.
   * @param right Right argument.
   * @return Switch result. 
   **/
  Result caseMultiplication(Term left, Term right);
  
  /**
   * Rational division.
   * @param left Dividend.
   * @param right Divisor.
   * @return Switch result. 
   **/
  Result caseDivision(Term left, Term right);
  
  /**
   * Remainder.
   * @param left Dividend.
   * @param right Divisor.
   * @return Switch result. 
   **/
  Result caseRemainder(Term left, Term right);
  
  /**
   * Exponentiation.
   * @param left Base.
   * @param right Exponent.
   * @return Switch result. 
   **/
  Result casePower(Term left, Term right);
    
  /**
   * Local variable binding.
   * @param var Variable name.
   * @param type Type of the bound variable.
   * @param value Assigned value.
   * @param body Expression body.
   * @return Switch result. 
   **/
  Result caseLet(String var,
                 ValueType type,  
                 Term value, 
                 Term body);
  
  /**
   * Local propositional variable binding.
   * @param var Variable name.
   * @param value Assigned value.
   * @param body Expression body.
   * @return Switch result. 
   **/
  Result caseLetProp(String var, Formula value, Term body);
  
  /**
   * If-then-else construct.
   * @param cond Condition.
   * @param ifTrue If part.
   * @param ifFalse Else part.
   * @return Switch result. 
   **/
  Result caseIfThenElse(Formula cond, Term ifTrue, Term ifFalse);
  
  /**
   * Array access.
   * @param arr Array.
   * @param index Index.
   * @return Switch result. 
   **/
  Result caseArrayAccess(Term arr, Term index);
  
  /**
   * Array update.
   * @param arr Array.
   * @param index Index.
   * @param value New value.
   * @return Switch result. 
   **/
  Result caseArrayUpdate(Term arr, Term index, Term value);
  
  /**
   * Function call.
   * @param func Function name.
   * @param args Arguments.
   * @return Switch result. 
   **/
  Result caseCall(String func, Term... args);
  
  /**
   * Bit vector concatenation.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseConcat(Term left, Term right);
  
  /**
   * Bit vector selection.
   * 
   * Select bits at positions left through right (inclusive).
   *  
   * @param vec Vector.
   * @param left Leftmost bit index.
   * @param right Rightmost bit index.
   * @return Switch result. 
   **/
  Result caseSelectBits(Term vec, int left, int right);
  
  /**
   * Shift left.
   * @param vec Vector.
   * @param offset Offset.
   * @return Switch result. 
   **/
  Result caseShl(Term vec, int offset);
  
  /**
   * Shift right.
   * @param vec Vector.
   * @param offset Offset.
   * @return Switch result. 
   **/
  Result caseShr(Term vec, int offset);
  
  /**
   * Sign extension.
   * @param vec Vector.
   * @param offset Number of added bits.
   * @return Switch result. 
   **/
  Result caseSX(Term vec, int offset);
  
  /**
   * Zero extension.
   * @param vec Vector.
   * @param offset Number of added bits.
   * @return Switch result. 
   **/
  Result caseZeroExtend(Term vec, int offset);
  
  /**
   * Repeat vector.
   * @param vec Vector.
   * @param count Number of copies to concatenate.
   * @return Switch result. 
   **/
  Result caseRepeat(Term vec, int count);

  /**
   * Rotate left.
   * @param vec Vector.
   * @param offset Offset.
   * @return Switch result. 
   **/
  Result caseRotateLeft(Term vec, int offset);
  
  /**
   * Rotate right.
   * @param vec Vector.
   * @param offset Offset.
   * @return Switch result. 
   **/
  Result caseRotateRight(Term vec, int offset);
  
  /**
   * Bitwise and.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorAnd(Term left, Term right);
 
  /**
   * Bitwise or.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorOr(Term left, Term right);
  
  /**
   * Bitwise xor.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorXor(Term left, Term right);
 
  /**
   * Bitwise nand.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorNand(Term left, Term right);
  
  /**
   * Bitwise nor.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorNor(Term left, Term right);
  
  /**
   * Bitwise xnor.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorXnor(Term left, Term right);
  
  /**
   * Bit vector comparison.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorCmp(Term left, Term right);
  
  /**
   * Vector addition.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorSum(Term left, Term right);
    
  /**
   * Vector subtraction.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorDiff(Term left, Term right);
  
  /**
   * Unary minus on vectors.
   * @param vec Vector.
   * @return Switch result. 
   **/
  Result caseVectorUminus(Term vec);
  
  /**
   * Vector multiplication.
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorProd(Term left, Term right);
  
  /**
   * Shift left with offset given as vector.
   * @param vec Vector.
   * @param offset Offset vector.
   * @return Switch result. 
   **/
  Result caseVectorShl(Term vec, Term offset);

  /**
   * Logical shift right with offset given as vector.
   * @param vec Vector.
   * @param offset Offset vector.
   * @return Switch result. 
   **/
  Result caseVectorLshr(Term vec, Term offset);
  
  /**
   * Arithmetic shift right with offset given as vector.
   * 
   * New bits are set according to the sign bit 
   * of the original vector.
   * 
   * @param vec Vector.
   * @param offset Offset vector.
   * @return Switch result. 
   **/
  Result caseVectorRshr(Term vec, Term offset);
  
  /**
   * Unsigned vector division.
   * @param left Dividend.
   * @param right Divisor.
   * @return Switch result. 
   **/
  Result caseVectorUdiv(Term left, Term right);
  
  /**
   * Signed vector division.
   * @param left Dividend.
   * @param right Divisor.
   * @return Switch result. 
   **/
  Result caseVectorSdiv(Term left, Term right);
  
  /**
   * Unsigned vector remainder.
   * @param left Dividend.
   * @param right Divisor.
   * @return Switch result. 
   **/
  Result caseVectorUrem(Term left, Term right);

  /**
   * Signed vector remainder (sign follows dividend).
   * @param left Dividend.
   * @param right Divisor.
   * @return Switch result. 
   **/
  Result caseVectorSrem(Term left, Term right);

  /**
   * Signed vector modulus (sign follows divisor).
   * @param left Dividend.
   * @param right Divisor.
   * @return Switch result. 
   **/
  Result caseVectorSmod(Term left, Term right);

  /**
   * Vector unsigned less than.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorLt(Term left, Term right);

  /**
   * Vector unsigned greater than.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorGt(Term left, Term right);

  /**
   * Vector unsigned less than or equal to.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorLe(Term left, Term right);

  /**
   * Vector unsigned greater than or equal to.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorGe(Term left, Term right);
  
  /**
   * Vector signed less than.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorSlt(Term left, Term right);

  /**
   * Vector signed greater than.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorSgt(Term left, Term right);

  /**
   * Vector signed less than or equal to.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorSle(Term left, Term right);

  /**
   * Vector signed greater than or equal to.
   * 
   * Produces a single-bit vector.
   * 
   * @param left Left vector.
   * @param right Right vector.
   * @return Switch result. 
   **/
  Result caseVectorSge(Term left, Term right);

}
