package net.afyre.haha.demo.solver.expr.util;

import java.io.IOException;
import java.io.Writer;
import java.math.BigInteger;

import net.afyre.haha.demo.solver.expr.Expression;
import net.afyre.haha.demo.solver.expr.ExpressionSwitch;
import net.afyre.haha.demo.solver.expr.VariableBinding;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.solver.expr.terms.Term;
import net.afyre.haha.demo.solver.expr.types.ValueType;

/**
 * Simple expression printer, using infix notation.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public class ExprPrinter implements ExpressionSwitch<Void> {
  /** Output stream. */
  private final Writer output;
  
  /** Switch used to calculate operator priorities. */
  private final PrioritySwitch prioritySwitch;
  
  /**
   * Create a visitor which will write all 
   * inspected objects to an output stream.
   * @param output Output stream.
   */
  public ExprPrinter(final Writer output) {
    this.output = output;
    this.prioritySwitch = new PrioritySwitch();
  }
  
  /**
   * Print an object (using toString()) to the output stream, 
   * converting any IOException to RuntimeExceptions.
   * @param obj Object to be printed (can be null).
   */
  private void print(final Object obj) {
    final String message;
    
    if (obj == null) {
      message = "null";
    } else {
      message = obj.toString();
    }
    try {
      output.write(message);
    } catch (final IOException e) {
      throw new RuntimeException(e);
    }
  }
  
  /**
   * Print the left subexpression of a binary expression, 
   * surrounding it with parentheses if necessary.
   *  
   * @param expr Expression to be printed
   * @param outerPriority Priority of the outer operator.
   */
  private void printLeft(final Expression expr, final int outerPriority) {
    final int priority = expr.accept(prioritySwitch).getLeft();
    if (priority < outerPriority) {
      print("(");
    }
    expr.accept(this);
    if (priority < outerPriority) {
      print(")");
    }
  }

  /**
   * Print the right subexpression of a binary expression, 
   * surrounding it with parentheses if necessary.
   *  
   * @param expr Expression to be printed
   * @param outerPriority Priority of the outer operator.
   */
  private void printRight(final Expression expr, final int outerPriority) {
    final int priority = expr.accept(prioritySwitch).getRight();
    if (priority < outerPriority) {
      print("(");
    }
    expr.accept(this);
    if (priority < outerPriority) {
      print(")");
    }
  }
  
  /**
   * Print a subexpression, surrounding it with parentheses if necessary.
   *  
   * @param expr Expression to be printed
   * @param outerPriority Priority of the outer operator.
   */
  private void printExpr(final Expression expr, final int outerPriority) {
    printLeft(expr, outerPriority);
  }
  
  /**
   * Print a list of subexpressions using given separator.
   * 
   * @param <T> Type of subexpressions (terms, formulae, ...).
   * @param exprs List of subexpressions.
   * @param sep Separator string.
   * @param sepPriority Priority of the separating operator.
   */
  private<T extends Expression> 
      void printList(final T[] exprs, 
                     final String sep, 
                     final int sepPriority) {
    if (exprs.length == 1) {
      exprs[0].accept(this);
    } else if (exprs.length > 1) {
      printLeft(exprs[0], sepPriority);
      for (int i = 1; i < exprs.length - 1; i++) {
        final Priority priority = exprs[i].accept(prioritySwitch);
        print(sep);
        if (priority.getLeft() < sepPriority || 
            priority.getRight() < sepPriority) {
          print("(");
        }
        exprs[i].accept(this);
        if (priority.getLeft() < sepPriority || 
            priority.getRight() < sepPriority) {
          print(")");
        }
      }
      print(sep);
      printRight(exprs[exprs.length - 1], sepPriority);
    }
  }
  
  /** {@inheritDoc} */
  @Override
  public Void caseIntConst(final BigInteger value) {
    print(value);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseFormulaVariable(final String var) {
    print(var);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVariable(final String var) {
    print(var);
    return null;
  }
  
  /** {@inheritDoc} */
  @Override
  public Void caseMkTuple(final Term... elements) {
    print("(");
    printList(elements, ", ", Priority.COMMA + 1);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseTupleAccess(final Term tuple, final int index) {
    printExpr(tuple, Priority.MAX);
    print("[");
    print(index);
    print("]");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseUminus(final Term arg) {
    print("-");
    printExpr(arg, Priority.MAX);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseAddition(final Term left, final Term right) {
    printLeft(left, Priority.SUM);
    print(" + ");
    printRight(right, Priority.SUM);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseSubtraction(final Term left, final Term right) {
    printLeft(left, Priority.SUM);
    print(" - ");
    printRight(right, Priority.SUM + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseMultiplication(final Term left, final Term right) {
    printLeft(left, Priority.PROD);
    print(" * ");
    printRight(right, Priority.PROD);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseDivision(final Term left, final Term right) {
    printLeft(left, Priority.PROD);
    print(" / ");
    printRight(right, Priority.PROD + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseLet(final String var, 
                      final ValueType type, 
                      final Term value, 
                      final Term body) {
    print("let ");
    print(var);
    print(" : ");
    type.accept(this);
    print(" = ");
    value.accept(this);
    print(" in ");
    body.accept(this);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseLetProp(final String var, 
                          final Formula value, 
                          final Term body) {
    print("let ");
    print(var);
    print(" = ");
    value.accept(this);
    print(" in ");
    body.accept(this);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseIfThenElse(final Formula cond, 
                              final Term ifTrue, 
                              final Term ifFalse) {
    print("if ");
    cond.accept(this);
    print(" then ");
    ifTrue.accept(this);
    print(" else ");
    ifFalse.accept(this);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseArrayAccess(final Term arr, 
                               final Term index) {
    printExpr(arr, Priority.MAX);
    print("[");
    index.accept(this);
    print("]");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseArrayUpdate(final Term arr, 
                              final Term index, 
                              final Term value) {
    printExpr(arr, Priority.MAX);
    print("[");
    index.accept(this);
    print("]");
    print(" := ");
    value.accept(this); 
    return null;  
  }

  /** {@inheritDoc} */
  @Override
  public Void caseCall(final String func, 
                       final Term... args) {
    print(func);
    print("(");
    printList(args, ", ", Priority.COMMA + 1);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void casePredicateCall(final String pred, 
                                final Term... args) {
    print(pred);
    print("(");
    printList(args, ", ", Priority.COMMA + 1);
    print(")");
    return null;
  }
  
  /** {@inheritDoc} */
  @Override
  public Void caseConcat(final Term left, final Term right) {
    print("concat");
    print("(");
    printLeft(left, Priority.COMMA + 1);
    print(", ");
    printRight(right, Priority.COMMA + 1);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseSelectBits(final Term vec, 
                             final int left, 
                             final int right) {
    printExpr(vec, Priority.MAX);
    print("[");
    print(left);
    print(":");
    print(right);
    print("]");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseShl(final Term vec, 
                      final int offset) {
    printLeft(vec, Priority.SHIFT + 1);
    print(" << ");
    print(offset);
    return null;  
  }

  /** {@inheritDoc} */
  @Override
  public Void caseShr(final Term vec, 
                      final int offset) {
    printLeft(vec, Priority.SHIFT + 1);
    print(" >> ");
    print(offset);
    return null;  
  }

  /** {@inheritDoc} */
  @Override
  public Void caseSX(final Term vec, 
                     final int offset) {
    print("(SIGN_EXTENSION ");
    vec.accept(this);
    print(" ");
    print(offset);
    print(")"); 
    return null;  
  }

  /** {@inheritDoc} */
  @Override
  public Void caseZeroExtend(final Term vec, 
                              final int offset) {
    print("(ZERO_EXTENSION ");
    vec.accept(this);
    print(" ");
    print(offset);
    print(")"); 
    return null;  
  }

  /** {@inheritDoc} */
  @Override
  public Void caseRepeat(final Term vec, 
                          final int count) {
    print("(REPEAT ");
    vec.accept(this);
    print(" ");
    print(count);
    print(")");
    return null;   
  }

  /** {@inheritDoc} */
  @Override
  public Void caseRotateLeft(final Term vec, 
                              final int offset) {
    print("(ROTATE_LEFT ");
    vec.accept(this);
    print(" ");
    print(offset);
    print(")");   
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseRotateRight(final Term vec, 
                               final int offset) {
    print("(ROTATE_RIGHT ");
    vec.accept(this);
    print(" ");
    print(offset);
    print(")");  
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorAnd(final Term left, 
                             final Term right) {
    print("(VECTOR_AND ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorOr(final Term left, 
                            final Term right) {
    print("(VECTOR_OR ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorXor(final Term left, 
                             final Term right) {
    print("(VECTOR_XOR ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorNand(final Term left, 
                              final Term right) {
    print("(VECTOR_NAND ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorNor(final Term left, 
                             final Term right) {
    print("(VECTOR_NOR ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorXnor(final Term left, 
                              final Term right) {
    print("(VECTOR_XNOR ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorCmp(final Term left, 
                             final Term right) {
    print("(VECTOR_CMP ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSum(final Term left, 
                             final Term right) {
    print("(VECTOR_ADD ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorDiff(final Term left, 
                              final Term right) {
    print("(VECTOR_SUB ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorUminus(final Term vec) {
    print("(VECTOR_UMINUS ");
    vec.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorProd(final Term left, 
                              final Term right) {
    print("(VECTOR_MUL ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorShl(final Term vec, 
                             final Term offset) {
    print("(VECTOR_SHL ");
    vec.accept(this);
    print(" ");
    offset.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorLshr(final Term vec, 
                              final Term offset) {
    print("(VECTOR_LSHR ");
    vec.accept(this);
    print(" ");
    offset.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorRshr(final Term vec, 
                              final Term offset) {
    print("(VECTOR_RSHR ");
    vec.accept(this);
    print(" ");
    offset.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorUdiv(final Term left, 
                              final Term right) {
    print("(VECTOR_UDIV ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSdiv(final Term left, 
                              final Term right) {
    print("(VECTOR_SDIV ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorUrem(final Term left, 
                              final Term right) {
    print("(VECTOR_UREM ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSrem(final Term left, 
                              final Term right) {
    print("(VECTOR_SREM ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSmod(final Term left, 
                              final Term right) {
    print("(VECTOR_SMOD ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorLt(final Term left, 
                            final Term right) {
    print("(VECTOR_LT ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorGt(final Term left, 
                            final Term right) {
    print("(VECTOR_GT ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorLe(final Term left, 
                            final Term right) {
    print("(VECTOR_LE ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorGe(final Term left, 
                            final Term right) {
    print("(VECTOR_GE ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSlt(final Term left, 
                             final Term right) {
    print("(VECTOR_SLT ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSgt(final Term left, 
                             final Term right) {
    print("(VECTOR_SGT ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSle(final Term left, 
                             final Term right) {
    print("(VECTOR_SLE ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseVectorSge(final Term left, 
                             final Term right) {
    print("(VECTOR_SGE ");
    left.accept(this);
    print(" ");
    right.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseTrue() {
    print("TRUE");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseFalse() {
    print("FALSE");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseNegation(final Formula arg) {
    print("¬");
    printExpr(arg, Priority.MAX);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseConjunction(final Formula left, 
                              final Formula right) {
    printLeft(left, Priority.AND);
    print(" ∧ ");
    printRight(right, Priority.AND);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseDisjunction(final Formula left, 
                              final Formula right) {
    printLeft(left, Priority.OR);
    print(" ∨ ");
    printRight(right, Priority.OR);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseImplication(final Formula left, 
                              final Formula right) {
    printLeft(left, Priority.IMPL + 1);
    print(" → ");
    printRight(right, Priority.IMPL);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseXor(final Formula left, 
                       final Formula right) {
    print("xor(");
    printLeft(left, Priority.COMMA + 1);
    print(", ");
    printRight(right, Priority.COMMA + 1);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseEquivalence(final Formula left, 
                               final Formula right) {
    printLeft(left, Priority.IFF + 1);
    print(" ↔ ");
    printRight(right, Priority.IFF + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseEq(final Term left, final Term right) {
    printLeft(left, Priority.CMP + 1);
    print(" = ");
    printRight(right, Priority.CMP + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseNeq(final Term left, final Term right) {
    printLeft(left, Priority.CMP + 1);
    print(" ≠ ");
    printRight(right, Priority.CMP + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseDistinct(final Term... args) {
    printList(args, " ≠ ", Priority.CMP + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseLt(final Term left, final Term right) {
    printLeft(left, Priority.CMP + 1);
    print(" < ");
    printRight(right, Priority.CMP + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseGt(final Term left, final Term right) {
    printLeft(left, Priority.CMP + 1);
    print(" > ");
    printRight(right, Priority.CMP + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseLe(final Term left, final Term right) {
    printLeft(left, Priority.CMP + 1);
    print(" ≤ ");
    printRight(right, Priority.CMP + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseGe(final Term left, final Term right) {
    printLeft(left, Priority.CMP + 1);
    print(" ≥ ");
    printRight(right, Priority.CMP + 1);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseExists(final Iterable<VariableBinding> variables,
                         final Formula arg,
                         final Expression... patterns) {
    printQuaintifiedFormula("∃", variables, arg, patterns);
    return null;
  }
  
  /** {@inheritDoc} */
  @Override
  public Void caseForall(final Iterable<VariableBinding> variables,
                         final Formula arg,
                         final Expression... patterns) {
    printQuaintifiedFormula("∀", variables, arg, patterns);
    return null;
  }

  /**
   * Helper function for printing quantified formulae.
   * @param quantifier Quantifier name.
   * @param variables Bound variables.
   * @param arg Formula body.
   * @param patterns Instantiation patterns.
   */
  private void printQuaintifiedFormula(
      final String quantifier,
      final Iterable<VariableBinding> variables,
      final Formula arg,
      final Expression... patterns) {

    print(quantifier);
    boolean first = true;
    for (final VariableBinding variable : variables) {
      if (!first) {
        print(", ");
      }
      print(variable.getName());
      print(" : ");
      variable.getType().accept(this);
      print(" ");
      first = false;
    }
    print(", ");
    printRight(arg, Priority.QUANT);
    if (patterns.length > 0) {
      print("patterns(");
      printList(patterns, ", ", Priority.COMMA + 1);
      print(")");
    }
  }
  
  /** {@inheritDoc} */
  @Override
  public Void caseFLet(final String var, 
                       final ValueType type,
                       final Term value, 
                       final Formula body) {
    print("let ");
    print(var);
    print(" : ");
    type.accept(this);
    print(" = ");
    value.accept(this);
    print(" in ");
    body.accept(this);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseFLetProp(final String var, 
                           final Formula value, 
                           final Formula body) {
    print("let ");
    print(var);
    print(" = ");
    value.accept(this);
    print(" in ");
    body.accept(this);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseRational() {
    print("REAL"); // As in CVC3
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseInt() {
    print("Z");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseBitvector(final int size) {
    print("INT" + size);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseArray(final ValueType indexType, 
                         final ValueType valueType) {
    print("array[");
    indexType.accept(this);
    print("] of ");
    valueType.accept(this);
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseTuple(final ValueType... types) {
    print("(TUPLE");
    for (final ValueType type : types) {
      print(" ");
      type.accept(this);
    }
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseProp() {
    print("PROP");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseFunction(final ValueType resultType, 
                            final ValueType... argTypes) {
    print("(FUNCTION ");
    for (final ValueType argType : argTypes) {
      argType.accept(this);
      print(" ");
    }
    resultType.accept(this);
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void casePredicate(final ValueType... argTypes) {
    print("(PREDICATE");
    for (final ValueType argType : argTypes) {
      print(" ");
      argType.accept(this);
    }
    print(")");
    return null;
  }

  /** {@inheritDoc} */
  @Override
  public Void caseRemainder(final Term left, final Term right) {
    printLeft(left, Priority.EXP);
    print(" % ");
    printRight(right, Priority.EXP + 1);
    return null; 
  }

  /** {@inheritDoc} */
  @Override
  public Void casePower(final Term left, final Term right) {
    printLeft(left, Priority.EXP + 1);
    print("^");
    printRight(right, Priority.EXP);
    return null;
  }
  
  private static final class Priority {
    public static final int MAX = 1000;
    public static final int EXP = 300;
    public static final int PROD = 200;
    public static final int SUM = 100;
    public static final int SHIFT = 70;
    public static final int CMP = 60;
    public static final int QUANT = 50;
    public static final int AND = 40;
    public static final int OR = 30;
    public static final int IMPL = 20;
    public static final int IFF = 10;
    public static final int COMMA = 5;
    public static final int MIN = 0;
    
    private final int left;
    private final int right;
    
    Priority(final int left, final int right) {
      this.left = left;
      this.right = right;
    }
    
    public int getLeft() {
      return left;
    }
    
    public int getRight() {
      return right;
    }
  }
  
  private static final class PrioritySwitch 
      extends DefaultExpressionSwitch<Priority> {
    private static final Priority PRIORITY_MAX = 
        new Priority(Priority.MAX, Priority.MAX);
    private static final Priority PRIORITY_EXP = 
        new Priority(Priority.EXP, Priority.EXP);
    private static final Priority PRIORITY_PROD = 
        new Priority(Priority.PROD, Priority.PROD);
    private static final Priority PRIORITY_SUM = 
        new Priority(Priority.SUM, Priority.SUM);
    private static final Priority PRIORITY_SHIFT = 
        new Priority(Priority.SHIFT, Priority.SHIFT);
    private static final Priority PRIORITY_CMP = 
        new Priority(Priority.CMP, Priority.CMP);
    private static final Priority PRIORITY_QUANT = 
        new Priority(Priority.QUANT, Priority.MAX);
    private static final Priority PRIORITY_AND = 
        new Priority(Priority.AND, Priority.AND);
    private static final Priority PRIORITY_OR = 
        new Priority(Priority.OR, Priority.OR);
    private static final Priority PRIORITY_IMPL = 
        new Priority(Priority.IMPL, Priority.IMPL);
    private static final Priority PRIORITY_IFF = 
        new Priority(Priority.IFF, Priority.IFF);
//    private static final Priority PRIORITY_MIN = 
//        new Priority(Priority.MIN, Priority.MIN);
    private static final Priority PRIORITY_LET = 
        new Priority(Priority.MIN, Priority.MAX);
    
    /** {@inheritDoc} */
    @Override
    protected Priority getDefault() {
      return PRIORITY_MAX;
    }
    /** {@inheritDoc} */
    @Override
    public Priority caseAddition(final Term left, final Term right) {
      return PRIORITY_SUM;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseSubtraction(final Term left, final Term right) {
      return PRIORITY_SUM;
    }

    @Override
    public Priority caseMultiplication(final Term left, final Term right) {
      return PRIORITY_PROD;
    }

    @Override
    public Priority caseDivision(final Term left, final Term right) {
      return PRIORITY_PROD;
    }

    @Override
    public Priority caseRemainder(final Term left, final Term right) {
      return PRIORITY_PROD;
    }

    @Override
    public Priority casePower(final Term left, final Term right) {
      return PRIORITY_EXP;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseLet(final String var, final ValueType type, 
                            final Term value, final Term body) {
      return PRIORITY_LET;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseLetProp(final String var, final Formula value, 
                                final Term body) {
      return PRIORITY_LET;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseIfThenElse(final Formula cond, final Term ifTrue, 
                                   final Term ifFalse) {
      return PRIORITY_LET;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseArrayUpdate(final Term arr, 
                                    final Term index, 
                                    final Term value) {
      return PRIORITY_LET;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseConjunction(final Formula left, final Formula right) {
      return PRIORITY_AND;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseDisjunction(final Formula left, final Formula right) {
      return PRIORITY_OR;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseImplication(final Formula left, final Formula right) {
      return PRIORITY_IMPL;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseEquivalence(final Formula left, final Formula right) {
      return PRIORITY_IFF;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseEq(final Term left, final Term right) {
      return PRIORITY_CMP;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseNeq(final Term left, final Term right) {
      return PRIORITY_CMP;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseLt(final Term left, final Term right) {
      return PRIORITY_CMP;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseGt(final Term left, final Term right) {
      return PRIORITY_CMP;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseLe(final Term left, final Term right) {
      return PRIORITY_CMP;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseGe(final Term left, final Term right) {
      return PRIORITY_CMP;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseExists(final Iterable<VariableBinding> variables, 
                               final Formula arg,
                               final Expression... patterns) {
      return PRIORITY_QUANT;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseForall(final Iterable<VariableBinding> variables, 
                               final Formula arg,
                               final Expression... patterns) {
      return PRIORITY_QUANT;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseFLet(final String var, final ValueType type, 
                             final Term value, final Formula body) {
      return PRIORITY_LET;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseFLetProp(final String var, final Formula value, 
                                 final Formula body) {
      return PRIORITY_LET;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseShl(final Term vec, final int offset) {
      return PRIORITY_SHIFT;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseShr(final Term vec, final int offset) {
      return PRIORITY_SHIFT;
    }

    /** {@inheritDoc} */
    @Override
    public Priority caseDistinct(final Term... args) {
      return PRIORITY_CMP;
    }
    
    
  }
}


