Require Export ZArith ZArith.Int.
Local Open Scope Z_scope.

Create HintDb haha_hints.

(** * Array notations and properties *)

(** Array type notation *)
Notation "'array' [ Y ]" := (Z -> Y) (at level 35, right associativity,
                                            format "'array' [ Y ]").

(** Array lookup *)

Definition lookup {Y} (A : array[Y]) i := (A i).

Notation "A [ i ]" := (lookup A i) (at level 5, left associativity).

(** Array update *)

Definition update {Y} (A : array[Y]) (where_ : Z) (val : Y) :=
		      fun (where__ : Z) =>  if (Z.eqb where_ where__) then val else A[where__].

Notation "A [ i <- v ]" := (update A i v) (at level 5, left associativity). 

Lemma update_here: forall {Y} A i (v:Y) j, i = j -> A[i <- v][j] = v.
Proof.
  intros.
  unfold update, lookup.
  apply Z.eqb_eq in H.
  now rewrite H.
Qed.

Lemma update_elsewhere: forall {Y} A i (v:Y) j, i <> j -> A[i <- v][j] = A[j].
Proof.
  intros.
  unfold update, lookup.
  apply Z.eqb_neq in H.
  now rewrite H.
Qed.

(********************************)
(** * Tactics *)
(********************************)


(** ** General helper tactics *)

Hint Extern 10 (_ = _) => f_equal.


(** The following two tactics are taken from 
   [https://www.cis.upenn.edu/~bcpierce/sf/current/LibTactics.html]
*)
Require BinPos Coq.ZArith.BinInt.

Definition ltac_nat_from_int (x:BinInt.Z) : nat :=
  match x with
  | BinInt.Z0 => 0%nat
  | BinInt.Zpos p => BinPos.nat_of_P p
  | BinInt.Zneg p => 0%nat
  end.

Ltac nat_from_number N :=
  match type of N with
  | nat => constr:(N)
  | BinInt.Z =>  let N' := constr:(ltac_nat_from_int N) in eval compute in N'
  end.


(** Smart repeat tactic [t] at most [n] times - with unary [n] *)

Ltac smart_repeat n t :=
  (* idtac n; *)
  match n with 
  | O => 
      idtac
  | (S ?m) => 
      (progress t; smart_repeat m t) || idtac
  end.

(** ... with unary or binary [N] *)
Ltac wrap_smart_repeat N t :=
  let n := nat_from_number N in
  smart_repeat n t.

(** Example: [rep 100 dosomething] *)

Tactic Notation "rep" constr(n) tactic(t) := wrap_smart_repeat n t.


(** For temporarily blocking some hypotheses *)

Definition HIDE (E:Prop) := E.

Ltac hide H :=
  let E := type of H in
  change (HIDE E) in H.


(********************************************)
(** ** Tactics to solve array updates *)


(** Simpifies [A[i <- v] [j]] everywhere *)

Ltac solve_this_update A i v j :=
  let H := fresh "H" in
  first [ rewrite !(update_elsewhere A i v j) in * |- * by auto 1 with zarith
        | rewrite !(update_here A i v j) in * |- * by auto 1 with zarith
        | (assert (i<>j) as H by omega;
            rewrite !(update_elsewhere A i v j) in * |- * by exact H)
        | (assert (i=j) as H by omega;
            rewrite !(update_here A i v j) in * |- * by exact H)
        ].


(** Two cases [i=j] and [i<>j] *)

Ltac split_this_update A i v j := destruct (Z.eq_dec i j).


(** Find "lookup update" and call [tac] on it *)

Ltac find_update tac :=
  match goal with 
  | [ |- context [ (update ?A ?i ?v)[?j] ] ] => 
    tac A i v j
  | [ hyp : context [ (update ?A ?i ?v)[?j] ] |- _ ] => 
    tac A i v j
  end.


(** Solve all updates either by simplifying or splitting *)

Ltac solve_updates :=
  rep 3   (* 2 should be enough, but just in case... *)
    (repeat find_update solve_this_update; (* solve all easy updates *)
     auto with zarith;                     (* clean up *)
     try find_update split_this_update).   (* try to split one "uneasy" update (if any) *)
     (* after splitting, an "uneasy" update becomes easy in both branches *)
     (* that is the primary reason for rep *)



(** ** Tactics to simplify equalities *)

Ltac is_pos_const t := 
  match t with 
    | xH => true
    | xO ?x => is_pos_const x
    | xI ?x => is_pos_const x
    | _ => false
  end.

Ltac is_const t :=
  match t with
    | Z0 => true
    | Zpos ?t => is_pos_const t
    | Zneg ?t => is_pos_const t
    | _ => false
  end.

(*
Ltac check_const t :=
  match is_const t with
  | true => idtac t; idtac "constant!"
  | false => idtac t; idtac "non-constant; fail"; fail 1
  end.

Ltac rewconst :=
  match goal with
    | [H: ?a = ?b |- _ ] => rewrite -> H in *
    | [H: ?a = ?b |- _ ] => rewrite <- H in *
  end.

Ltac instconst :=
  match goal with
    | [H: ?a = ?b |- _ ] => rewrite -> H in *
    | [H: ?a = ?b |- _ ] => rewrite <- H in *
    | [H: ?a = ?b |- _ ] => rewrite H in *
  end
  ||
  ring_simplify.


Ltac instconstn :=
  match goal with
    | [H: ?a = ?b |- _ ] => rewrite -> H in *
    | [H: ?a = ?b |- _ ] => rewrite <- H in *
    | [H: ?a = ?b |- _ ] => rewrite H in *
    | [H: ?a = ?b |- _ ] => rewrite H in *
  end
  ||
  ring_simplify.

 *)

(** Size of term: constants are -1, operations +2 *)
Ltac size t :=
  match is_const t with
  | true => constr:(-1)
  | false => 
      match t with
      | ?X ?Y => 
          let xs := size X in 
          let ys := size Y in 
          let xss := constr:xs in
          let yss := constr:ys in
          constr:(xs + ys)
      | _ => constr:(2) (* atom *)
      end
  end.

(** Size, but atoms are 100, to rewrite i -> expression *)
Ltac ssize t :=
  match t with
    | Z0 => constr:(-1)
    | ?X ?Y => size t
    | _ => constr:(100) (* just atom - a variable ? *)
  end.

(*
Ltac rew_to_smaller := 
  match goal with
    | [H: ?a = ?b |- _ ] => 
      let asss := ssize a in (* *) (* *)
      let ass := eval simpl in asss in (* *) (* *) 
      let bsss := ssize b in (* *)
      let bss := eval simpl in bsss in (* *) (* *)
      let cmp := eval compute in (Z.compare ass bss) in (* *) (* *)
      match cmp with 
      | Lt => rewrite <- H in *; change (HIDE (a=b)) in H
      | Gt => rewrite -> H in *; change (HIDE (a=b)) in H
      | Eq => rewrite -> H in *; change (HIDE (a=b)) in H
      end
  end.

Ltac hide H :=
  let E := type of H in
  change (HIDE E) in H.

Ltac rseq :=
  match goal with
    | [H: ?a = ?b |- _ ] => ring_simplify in H; hide H
  end.

Ltac simplify := repeat rew_to_smaller; unfold HIDE in *; repeat rseq; unfold HIDE in *; ring_simplify.
*)


(** Rewrite equalities exp1 = exp2 in towards the side of smaller size,
    and i = exp towards exp *) 
Ltac rew_to_smaller H := 
  let E := type of H in
  match E with ?a = ?b =>
      match b with context [a] => rewrite <- H in *; hide H end || 
      match a with context [b] => rewrite -> H in *; hide H end || 
      let asss := ssize a in (* *) (* *)
      let ass := eval simpl in asss in (* *) (* *) 
      let bsss := ssize b in (* *)
      let bss := eval simpl in bsss in (* *) (* *)
      let cmp := eval compute in (Z.compare ass bss) in (* *) (* *)
      match cmp with 
      | Lt => rewrite <- H in *; hide H
      | Gt => rewrite -> H in *; hide H
      | Eq => rewrite -> H in *; hide H
      end
  end.

(** Do ring_simplify and call tac on all equalities *)
Ltac ring_simplify_iter_hyps tac :=
  repeat match goal with
    | [H: _ = _ |- _ ] => ring_simplify in H; tac H
  end.

(** Simplify and use all equalities *)
Ltac simplify :=
  ring_simplify_iter_hyps rew_to_smaller;
  unfold HIDE in *;
  try ring_simplify.


(** ** Forward reasonning on integers *)

(** Transform all inequalities into a < b *)
Module LT.

Lemma lelt : forall n m, n <= m <-> n < m+1.
  intuition.
Qed.

Lemma gtlt : forall n m, m > n <-> n < m.
  intuition.
Qed.

Lemma gelt : forall n m, m >= n <-> n < m+1.
  intuition.
Qed.

Lemma nlelt : forall n m, ~ m <= n <-> n < m.
  intuition.
Qed.

Lemma ngelt : forall n m, ~ n >= m <-> n < m.
  intuition.
Qed.

Lemma ngtlt : forall n m, ~ n > m <-> n < m+1.
  intuition.
Qed.

Lemma nltlt : forall n m, ~ m < n <-> n < m+1.
  intuition.
Qed.

Lemma neqlt : forall n m, n <> m  <->  n < m \/ m < n.
  intros; omega.
Qed.

Lemma nnltlt : forall m n, ~ m < n+1 <-> n < m.
  intuition.
Qed.

Lemma nnltgt : forall m n, ~ m < n+1 <-> m > n.
  intuition.
Qed.

Lemma nnltle : forall m n, ~ m < n <-> n <= m.
  intuition.
Qed.

Lemma nnltge : forall m n, ~ m < n <-> m >= n.
  intuition.
Qed.

Lemma nltnle : forall m n, ~ m < n+1 <-> ~ m <= n.
  intuition.
Qed.

Lemma nltnge : forall m n, ~ m < n+1 <-> ~ n >= m.
  intuition.
Qed.

Lemma nltngt : forall m n, ~ m < n <-> ~ n > m.
  intuition.
Qed.

End LT.

Ltac ltify_ineq :=
  repeat match goal with
  | [ H : ?P -> False |- _ ] => fold (not P) in H
  | [ |- ?P -> False ] => fold (not P)
  | [ H : ?a <= ?c |- _ ] => rewrite -> LT.lelt in H; ring_simplify in H
  | [ H : ?a >= ?c |- _ ] => rewrite -> LT.gelt in H; ring_simplify in H
  | [ H : ?a > ?c |- _ ] => rewrite -> LT.gtlt in H; ring_simplify in H
  | [ |- ?a <= ?c ] => rewrite -> LT.lelt; ring_simplify
  | [ |- ?a >= ?c ] => rewrite -> LT.gelt; ring_simplify
  | [ |- ?a > ?c ] => rewrite -> LT.gtlt; ring_simplify
  | [ H : ~ ?a <= ?c |- _ ] => rewrite -> LT.nlelt in H; ring_simplify in H
  | [ H : ~ ?a >= ?c |- _ ] => rewrite -> LT.ngelt in H; ring_simplify in H
  | [ H : ~ ?a > ?c |- _ ] => rewrite -> LT.ngtlt in H; ring_simplify in H
  | [ H : ~ ?a < ?c |- _ ] => rewrite -> LT.nltlt in H; ring_simplify in H
  | [ |- ~ ?a <= ?c ] => rewrite -> LT.nlelt; ring_simplify
  | [ |- ~ ?a >= ?c ] => rewrite -> LT.ngelt; ring_simplify
  | [ |- ~ ?a > ?c ] => rewrite -> LT.ngtlt; ring_simplify
  | [ |- ~ ?a < ?c ] => rewrite -> LT.nltlt; ring_simplify
  | [ H : ?a <> ?c |- _ ] => apply -> LT.neqlt in H; destruct H; try omega
  end.

Ltac ltify_ineq_hyps :=
  repeat match goal with
  | [ H : ?P -> False |- _ ] => fold (not P) in H
  | [ |- ?P -> False ] => intro
  | [ H : ?a <= ?c |- _ ] => rewrite -> LT.lelt in H; ring_simplify in H
  | [ H : ?a >= ?c |- _ ] => rewrite -> LT.gelt in H; ring_simplify in H
  | [ H : ?a > ?c |- _ ] => rewrite -> LT.gtlt in H; ring_simplify in H
  | [ |- ?a <= ?c ] => rewrite <- LT.nnltle; let H := fresh in intro H; ring_simplify in H
  | [ |- ?a >= ?c ] => rewrite <- LT.nnltge; let H := fresh in intro H; ring_simplify in H
  | [ |- ?a > ?c ] => rewrite <- LT.nnltgt; let H := fresh in intro H; ring_simplify in H
  | [ |- ?a < ?c ] => rewrite <- LT.nnltlt; let H := fresh in intro H; ring_simplify in H
  | [ H : ~ ?a <= ?c |- _ ] => rewrite -> LT.nlelt in H; ring_simplify in H
  | [ H : ~ ?a >= ?c |- _ ] => rewrite -> LT.ngelt in H; ring_simplify in H
  | [ H : ~ ?a > ?c |- _ ] => rewrite -> LT.ngtlt in H; ring_simplify in H
  | [ H : ~ ?a < ?c |- _ ] => rewrite -> LT.nltlt in H; ring_simplify in H
  | [ |- ~ ?a <= ?c ] => rewrite <- LT.nltnle; let H := fresh in intro H; ring_simplify in H
  | [ |- ~ ?a >= ?c ] => rewrite <- LT.nltnge; let H := fresh in intro H; ring_simplify in H
  | [ |- ~ ?a > ?c ] => rewrite <- LT.nltngt; let H := fresh in intro H; ring_simplify in H
  | [ |- ~ ?a < ?c ] => let H := fresh in intro H; ring_simplify in H
  | [ H : ?a <> ?c |- _ ] => apply -> LT.neqlt in H; destruct H; try omega
  end.

(*

Module Test_ltify_ineq.

Lemma lelt : forall n m, n <= m <-> n < m+1.
intros; split; intros.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Qed.

Lemma gtlt : forall n m, m > n <-> n < m.
intros; split; intros.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Qed.

Lemma gelt : forall n m, m >= n <-> n < m+1.
intros; split; intros.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Qed.

Lemma nlelt : forall n m, ~ m <= n <-> n < m.
intros; split; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Qed.

Lemma ngelt : forall n m, ~ n >= m <-> n < m.
intros; split; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Qed.

Lemma ngtlt : forall n m, ~ n > m <-> n < m+1.
intros; split; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Qed.

Lemma nltlt : forall n m, ~ m < n <-> n < m+1.
intros; split; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq. trivial.
ltify_ineq. trivial.
Qed.

Lemma nnltlt : forall m n, ~ m < n+1 <-> n < m.
intros; split; intro.
Show 2.
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Qed.

Lemma nnltgt : forall m n, ~ m < n+1 <-> m > n.
intros; split; intro.
Show 2.
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Qed.

Lemma nnltle : forall m n, ~ m < n <-> n <= m.
intros; split; intro.
Show 2.
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Qed.

Lemma nnltge : forall m n, ~ m < n <-> m >= n.
intros; split; intro.
Show 2.
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Qed.

Lemma nltnle : forall m n, ~ m < n+1 <-> ~ m <= n.
intros; split; intro.
Show 2.
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Qed.

Lemma nltnge : forall m n, ~ m < n+1 <-> ~ n >= m.
intros; split; intro.
Show 2.
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Qed.

Lemma nltngt : forall m n, ~ m < n <-> ~ n > m.
intros; split; intro.
Show 2.
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Restart.
intros; split; unfold not; intro.
Show 2. 
ltify_ineq_hyps. omega.
ltify_ineq_hyps. omega.
Qed.


End Test_ltify_ineq.

*)


(** Transform [i - a] into [i + (-a)] in inequalities so we can work with [i + b] only *) 
Ltac replace_minus :=
  repeat match goal with
  | [ H : (?i - ?a) < ?b |- _ ] => replace (i - a) with (i + (-a)) in H by ring
  | [ H : ?b < (?i - ?a) |- _ ] => replace (i - a) with (i + (-a)) in H by ring
  end.

(*
(* bezużyteczne i chyba źle *)
Ltac ring_simp exp e :=
  let G := fresh "G" in
  evar (exp : Z);
  assert (exp = e) as G by (ring_simplify; unfold exp; reflexivity);
  clear G.
 *)

(** Replace type of hypothesis [H] by its omega-equivalent [t] 
   and then simplify [e] in [t] *)
Ltac replace_hyp H t e :=
  let G := fresh "G" in
  let t1 := type of H in
  assert (t1 -> t) as G by omega;
  apply G in H;
  clear G;
  ring_simplify e in H;
  hide H.

(** Transform all inequalities on [i] into [i < j + exp] or [j + exp < i] *)
Ltac one_var i :=
  repeat match goal with
  | [ H : i+?a < ?j+?b |- _ ] => 
      (* idtac H ": [" i "] +" a "<" j "+" b; *)
      replace_hyp H (i < j + (b-a)) (b-a)
  | [ H : i+?a < ?j |- _ ] => 
      (* idtac H ": [" i "] +" a "<" j;  *)
      replace_hyp H (i < j + (-a)) (-a)
  | [ H : i < ?j + _ |- _ ] =>
      hide H
  | [ H : ?j+?a < i+?b |- _ ] => 
      (* idtac H ":" j "+" a "< [" i "] +" b; *) 
      replace_hyp H (j + (a-b) < i) (a-b)
  | [ H : ?j < i+?b |- _ ] => 
      (* idtac H ":" j "< [" i "] +" b; *)
      replace_hyp H (j + (-b) < i) (-b)
  | [ H : ?j + _ < i |- _ ] =>
      hide H
  | [ H : i < ?j |- _ ] => 
      (* is_var j; (* this works only for local or section variables *) *)
      (* idtac H ": [" i "] <" j; *)
      replace_hyp H (i < j + 0) 0
  | [ H : ?j < i |- _ ] => 
      (* is_var j; (* this works only for local or section variables *) *)
      (* idtac H ":" j "< [" i "]";  *)
      replace_hyp H (j + 0 < i) 0
  end;
  unfold HIDE in *.

(* TODO: constants, i.e. i < c and c < j *)


(** Find [i < j + b] and [j + a < i], check if [a] and [b] are close enough 
    deduce suitable equality/ies and run [proceed] on them *)
Ltac deduce_eq i proceed :=
  match goal with
  | [ H1 : i < ?j + ?b, H2 : ?j + ?a < i |- _ ] => 
      (* idtac j ":" b "-" a;  *)
      let d := eval compute in (b-a) in
      (* idtac "Jest " d; *)
      let H := fresh "H" in
      match d with
      | Zneg _ => omega
      | 0 => omega
      | 1 => omega
      | 2 => assert (i=j+(a+1)) as H by omega;
            try omega;
            proceed H
      | 3 => assert (i=j+(a+1) \/ i=j+(a+2)) as H by omega;
            ring_simplify (a+1) (a+2) in H;
            destruct H as [H|H]; try omega; proceed H
      (* can be extended if necessary *)
      end
  end.

(** Deduce equality on i from inequalities and use it *)
Ltac deduce_equality i :=
  ltify_ineq_hyps;
(*  try ( *)
    let rew E := 
    ring_simplify in E;
    rewrite -> E in *;
    ring_simplify_iter_hyps hide; unfold HIDE in *;
    try ring_simplify;
    move E at bottom
  in
  replace_minus;
  one_var i;
  deduce_eq i rew.


(** Just deduce the equality on i from inequalities *)
Ltac debug_deduce_equality i := 
  ltify_ineq_hyps;
  replace_minus;
  one_var i;
  let ringsimp E := 
    ring_simplify_iter_hyps hide; unfold HIDE in *;
    (* rewrite -> E in *; *) 
    try ring_simplify;
    move E at bottom
  in  
  deduce_eq i ringsimp.


(* Tests

Parameter f : Z -> Z.

Goal forall i k l, j < i -> i > l -> i < j+2 -> i < l+4 -> i > k -> i < k+3 -> i < k+2 -> f j= f k.
intros.
repeat match goal with [ a : Z |- _ ] => deduce_equality a; auto with zarith end.
Restart.
intros.
deduce_equality j.
deduce_equality i.
auto with zarith.
Qed.


Goal forall i k, ~ k-3 >= i-5 -> i-2 < k+3 -> i <= k+3 -> i=k.
intros.
ltify_ineq.
Abort.

*)


Goal forall x y, x > y -> x <> y.
intros.
auto 1 with zarith.  
Show Proof.
Qed.

Print Unnamed_thm.

(** For [f x <= f y] prove case [x=y] automatically, try to prove remaining cases by omega *) 
Ltac fequal_ineq := 
  match goal with 
  | [ |- (?f ?x <= ?f ?y) ] => 
(*      idtac f x y;  *)
      let H:=fresh in 
      (assert (x<>y) by auto 1 with zarith; fail 2)
      || (assert (x=y \/ x<>y) as [H|H] by omega; 
          [rewrite H; auto with zarith | idtac]; 
          apply -> LT.neqlt in H;
          destruct H as [H|H]; try omega)
  | [ |- (?f ?x >= ?f ?y) ] => 
(*      idtac f x y;  *)
      let H:=fresh in 
      (assert (x<>y) by auto 1 with zarith; fail 2)
      || (assert (x=y \/ x<>y) as [H|H] by omega; 
          [rewrite H; auto with zarith | idtac]; 
          apply -> LT.neqlt in H;
          destruct H as [H|H]; try omega)
  end.


(* Test 
Parameter B: array[Z].

Goal (forall n m, n < m -> B[n] >= B[m]) -> forall n m, n < m+1 -> B[n] >= B[m].
  intros.
  fequal_ineq.
  
  deduce_equality n.
  auto with zarith.
Qed.

*)

Ltac ltlt tac :=
    match goal with
      [ _ : ?a < ?c |- ?a < ?b ] => tac a b c
    end.



Ltac limit_from_hyp :=

  let en' := fresh "en" in
  evar (en' : Z);
  let en := eval unfold en' in en' in

  let em' := fresh "em" in
  evar (em' : Z);
  let em := eval unfold em' in em' in

  let instenem a b c := 
         (* idtac "inst" a b c; *)
         instantiate (1:=a) in (Value of en');
         instantiate (1:=b) in (Value of em')
         (* assert (en = a) by reflexivity; *)
         (* assert (em = b) by reflexivity *)
  in
  let myH := fresh in
  match goal with [ |- ?G ] => assert (HIDE False -> G) as myH ; [
    intro;
    match goal with [ H : _ |- _ ] => apply H end;
    intuition;
    try omega;
    auto with zarith;
    ltify_ineq;
    ltlt instenem;
    contradiction
  | idtac];
  let en := eval unfold en' in en' in (* old en em look invalid... *)                           
  let em := eval unfold em' in em' in
  (*idtac en em;*)
  ((is_evar en || is_evar em); fail 1 "Unresolved evars!") || (clear myH;
(* evars must be instantiated! *)
  ((assert (en < em \/ en = em) as [?|?] by omega)||
   (assert (en < em \/ en = em \/ en = em + 1) as [?|[?|?]] by omega)||
    assert (en < em \/ en >= em) as [?|?] by omega) ; 
    [ auto with zarith; try omega | clear en' em' .. ])
  end.




(* Test 
Parameter B: array[Z].


(* m -> m+1, żeby zobaczyć evar *)
Goal (forall n m, 0 < n < m -> B[n] <= B[m]) -> forall n m, 0 < n < m+1 -> B[n] <= B[m].
  intros.
  intuition.
  limit_from_hyp.
  fequal_ineq.

(*  show limit_from_hyp step by step *)
  let en' := fresh "en" in
  evar (en' : Z);
  let en := eval unfold en' in en' in

  let em' := fresh "em" in
  evar (em' : Z);
  let em := eval unfold em' in em' in

  let myH := fresh in
  match goal with [ |- ?G ] => assert (HIDE False -> G) as myH end . 
    intro.
    match goal with [ H : _ |- _ ] => apply H end.
    intuition.

    instantiate (1:=n0) in (Value of en).  
    instantiate (1:=m-1) in (Value of em).  

    contradiction.


  clear H0;
  ((assert (en < em \/ en = em) as [?|?] by omega)||
   (assert (en < em \/ en = em \/ en = em + 1) as [?|[?|?]] by omega)||
    assert (en < em \/ en >= em) as [?|?] by omega) ; subst en em.
Admitted.  
*)


(** ** Sledgehammer tactic *)

Ltac haha_solve :=
  repeat autounfold with haha_predicates;
  intuition;
  (rep 5 simplify);
  solve_updates;
  try omega;
  auto with zarith;
  try fequal_ineq;
  try simplify;
  try omega;
  auto with zarith;
  try limit_from_hyp;
  try simplify;
  try omega;
  auto with zarith;
  auto with haha_hints;
(*  idtac "Jestem"; *)
  repeat match goal with [ i : Z |- _ ] => deduce_equality i end;
  try simplify;
  try omega;
  auto with zarith;
  idtac.
