package net.afyre.haha.demo.util.tests;

import static org.junit.Assert.*;
import static net.afyre.haha.demo.util.tests.Util.*;

import java.util.ArrayList;
import java.util.Iterator;

import net.afyre.haha.demo.util.OffsetIterable;

import org.junit.Test;

/**
 * Tests for the {@link OffsetIterable} class.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public class OffsetIterableTest {

  /** GetInstance does not return null. */
  @Test
  public void testInstanceNotNull() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 1);
    assertNotNull("GetInstance() returned null.", iterable);
  }

  /** GetInstance throws exception when called with null argument. */
  @Test(expected=NullPointerException.class)
  public void testInstanceNullException() {
    OffsetIterable.getInstance(null, 1);
  }
  
  /** GetInstance throws exception when called with negative offset. */
  @Test(expected=IllegalArgumentException.class)
  public void testNegativeOffset() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    OffsetIterable.getInstance(source, -1);    
  }

  /** Returned iterator is not null. */
  @Test
  public void testIteratorNotNull() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 1);
    assertNotNull("Null iterator returned", iterable.iterator());
  }
  
  /** All elements are returned for offset equal to zero. */
  @Test
  public void testZeroOffset() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    source.add(2);
    source.add(3);
    source.add(4);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 0);
    assertIteratorReturns(iterable.iterator(), 1, 2, 3, 4);
  }
  
  /** Elements are skipped for offset greater than zero. */
  @Test
  public void testPositiveOffset() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    source.add(2);
    source.add(3);
    source.add(4);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 2);
    assertIteratorReturns(iterable.iterator(), 3, 4);
  }
    
  /** No elements returned for offset equal to length. */
  @Test
  public void testOffsetEqLength() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    source.add(2);
    source.add(3);
    source.add(4);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 4);
    assertIteratorReturns(iterable.iterator());
  }  

  /** No elements returned for offset greater than length. */
  @Test
  public void testOffsetGtLength() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    source.add(2);
    source.add(3);
    source.add(4);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 100);
    assertIteratorReturns(iterable.iterator());
  }
  
  /**
   * Calling remove() before next() is called causes an exception.
   */
  @Test(expected=IllegalStateException.class)
  public void testRemoveBeforeNext() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    source.add(2);
    source.add(3);
    source.add(4);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 2);
    final Iterator<Integer> it = iterable.iterator();
    it.remove();
  }
  
  /** Remove after next is forwarded to the source iterator. */
  @Test
  public void testRemoveAfterNext() {
    final ArrayList<Integer> source = new ArrayList<Integer>();
    source.add(1);
    source.add(2);
    source.add(3);
    source.add(4);
    source.add(5);
    source.add(6);
    final Iterable<Integer> iterable = 
        OffsetIterable.getInstance(source, 2);
    final Iterator<Integer> it = iterable.iterator();
    it.next();
    it.next();
    it.remove();
    assertArrayEquals("Wrong array state after remove()", 
                      new Integer[] { 1, 2, 3, /*4,*/ 5, 6}, 
                      source.toArray(new Integer[]{}));
  }
}
