package net.afyre.haha.demo.util.tests;

import net.afyre.haha.demo.util.Pair;

import org.junit.Test;

import static org.junit.Assert.*;

/**
 * Tests for the {@link Pair} class.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public class PairTest {
  /** GetInstance does not return null. */
  @Test
  public void testInstanceNotNull() {
    Pair<?, ?> pair;
    
    pair = Pair.getInstance(1, 2);
    assertNotNull("Null returned from getInstance()", pair);
    pair = Pair.getInstance(null, 2);
    assertNotNull("Null returned from getInstance()", pair);
    pair = Pair.getInstance(1, null);
    assertNotNull("Null returned from getInstance()", pair);
    pair = Pair.getInstance(null, null);
    assertNotNull("Null returned from getInstance()", pair);
  }

  /** GetInstance does not fail for null first argument. */
  @Test
  public void testInstanceLeftNull() {
    final Pair<?, ?> pair = Pair.getInstance(null, 2);
    assertNotNull("Null returned from getInstance()", pair);
    assertEquals("Wrong first element", null, pair.getFirst());
    assertEquals("Wrong second element", 2, pair.getSecond());
  }
 
  /** GetInstance does not fail for null second argument. */
  @Test
  public void testInstanceRightNull() {
    final Pair<?, ?> pair = Pair.getInstance(1, null);
    assertNotNull("Null returned from getInstance()", pair);
    assertEquals("Wrong first element", 1, pair.getFirst());
    assertEquals("Wrong second element", null, pair.getSecond());
  }
  
  /** GetInstance does not fail for null arguments. */
  @Test
  public void testInstanceBothNull() {
    final Pair<?, ?> pair = Pair.getInstance(null, null);
    assertNotNull("Null returned from getInstance()", pair);
    assertEquals("Wrong first element", null, pair.getFirst());
    assertEquals("Wrong second element", null, pair.getSecond());
  }
  
  /** Left element is returned correctly. */
  @Test
  public void testLeft() {
    final Pair<?, ?> pair = Pair.getInstance(1, 2);
    assertEquals("Wrong first element", 1, pair.getFirst());
  }

  
  /** Right element is returned correctly. */
  @Test
  public void testRight() {
    final Pair<?, ?> pair = Pair.getInstance(1, 2);
    assertEquals("Wrong second element", 2, pair.getSecond());
  }
  
  /** Equals is reflexive. */
  @Test
  public void testEqualsReflexive() {
    Pair<?, ?> pair = Pair.getInstance(1, 2);
    assertTrue("Equals not reflexive", pair.equals(pair));
    pair = Pair.getInstance(null, 2);
    assertTrue("Equals not reflexive (left null)", pair.equals(pair));
    pair = Pair.getInstance(1, null);
    assertTrue("Equals not reflexive (right null)", pair.equals(pair));
    pair = Pair.getInstance(null, null);
    assertTrue("Equals not reflexive (pair of nulls)", pair.equals(pair));
  }
  
  /** Equals works correctly on equal pairs of non-null elements. */
  @Test
  public void testStructuralEquality() {
    final Pair<?, ?> pair1 = Pair.getInstance(1, 2);
    final Pair<?, ?> pair2 = Pair.getInstance(1, 2);
    assertTrue("Wrong comparison result", pair1.equals(pair2));
    assertTrue("Wrong comparison result", pair2.equals(pair1));
  }

  /** Equals works correctly on unequal pairs of non-null elements. */
  @Test
  public void testStructuralInequality() {
    final Pair<?, ?> pair1 = Pair.getInstance(1, 2);
    final Pair<?, ?> pair2 = Pair.getInstance(3, 4);
    assertFalse("Wrong comparison result", pair1.equals(pair2));
    assertFalse("Wrong comparison result", pair2.equals(pair1));
  }
  
  /** Equals is false on pairs which differ only in the first component. */
  @Test
  public void testStructuralInequalityLeftEq() {
    final Pair<?, ?> pair1 = Pair.getInstance(1, 3);
    final Pair<?, ?> pair2 = Pair.getInstance(2, 3);
    assertFalse("Wrong comparison result", pair1.equals(pair2));
    assertFalse("Wrong comparison result", pair2.equals(pair1));
  }
  
  /** Equals is false on pairs which differ only in the second component. */
  @Test
  public void testStructuralInequalityRightEq() {
    final Pair<?, ?> pair1 = Pair.getInstance(1, 2);
    final Pair<?, ?> pair2 = Pair.getInstance(1, 3);
    assertFalse("Wrong comparison result", pair1.equals(pair2));
    assertFalse("Wrong comparison result", pair2.equals(pair1));
  }
  
  /** Equals works correctly on pairs of nulls. */
  @Test
  public void testEqualsNullElements() {
    final Pair<?, ?> pair1 = Pair.getInstance(null, null);
    final Pair<?, ?> pair2 = Pair.getInstance(null, null);
    final Pair<?, ?> pair3 = Pair.getInstance(null, 2);
    final Pair<?, ?> pair4 = Pair.getInstance(1, null);
    final Pair<?, ?> pair5 = Pair.getInstance(1, 2);
    assertTrue("Wrong comparison result", pair1.equals(pair2));
    assertFalse("Wrong comparison result", pair1.equals(pair3));
    assertFalse("Wrong comparison result", pair1.equals(pair4));
    assertFalse("Wrong comparison result", pair1.equals(pair5));  
    assertTrue("Wrong comparison result", pair2.equals(pair1));
    assertFalse("Wrong comparison result", pair3.equals(pair1));
    assertFalse("Wrong comparison result", pair4.equals(pair1));
    assertFalse("Wrong comparison result", pair5.equals(pair1));   
  }
  
  /** Equals works correctly when the first element is null. */
  @Test
  public void testEqualsLeftNull() {
    final Pair<?, ?> pair1 = Pair.getInstance(null, 2);
    final Pair<?, ?> pair2 = Pair.getInstance(null, 2);
    final Pair<?, ?> pair3 = Pair.getInstance(1, 2);
    assertTrue("Wrong comparison result", pair1.equals(pair2));
    assertFalse("Wrong comparison result", pair1.equals(pair3)); 
    assertTrue("Wrong comparison result", pair2.equals(pair1));
    assertFalse("Wrong comparison result", pair3.equals(pair1));
  }

  
  /** Equals works correctly when the second element is null. */
  @Test
  public void testEqualsRightNull() {
    final Pair<?, ?> pair1 = Pair.getInstance(1, null);
    final Pair<?, ?> pair2 = Pair.getInstance(1, null);
    final Pair<?, ?> pair3 = Pair.getInstance(1, 2);
    assertTrue("Wrong comparison result", pair1.equals(pair2));
    assertFalse("Wrong comparison result", pair1.equals(pair3)); 
    assertTrue("Wrong comparison result", pair2.equals(pair1));
    assertFalse("Wrong comparison result", pair3.equals(pair1));
  }
  
  /** Equals returns false for null argument. */
  @Test
  public void testEqualsNull() {
    final Pair<?, ?> pair1 = Pair.getInstance(null, null);
    final Pair<?, ?> pair2 = Pair.getInstance(null, 2);
    final Pair<?, ?> pair3 = Pair.getInstance(1, null);
    final Pair<?, ?> pair4 = Pair.getInstance(1, 2);
    
    assertFalse("Wrong comparison result", pair1.equals(null)); 
    assertFalse("Wrong comparison result", pair2.equals(null)); 
    assertFalse("Wrong comparison result", pair3.equals(null));
    assertFalse("Wrong comparison result", pair4.equals(null));  
  }
  
  /** Equals returns false for argument which is not a pair. */
  @Test
  public void testEqualsWrongType() {
    final Pair<?, ?> pair1 = Pair.getInstance(null, null);
    final Pair<?, ?> pair2 = Pair.getInstance(null, 2);
    final Pair<?, ?> pair3 = Pair.getInstance(1, null);
    final Pair<?, ?> pair4 = Pair.getInstance(1, 2);
    final Object other = "1, 2";
    
    assertFalse("Wrong comparison result", pair1.equals(other)); 
    assertFalse("Wrong comparison result", pair2.equals(other)); 
    assertFalse("Wrong comparison result", pair3.equals(other));
    assertFalse("Wrong comparison result", pair4.equals(other));  
  }
  
  /** Hashcodes for structurally equal pairs re equal. */
  @Test
  public void testHashcode() {
    final Pair<?, ?> pair1a = Pair.getInstance(null, null);
    final Pair<?, ?> pair1b = Pair.getInstance(null, null);
    
    final Pair<?, ?> pair2a = Pair.getInstance(null, 2);
    final Pair<?, ?> pair2b = Pair.getInstance(null, 2);
    
    final Pair<?, ?> pair3a = Pair.getInstance(1, null);
    final Pair<?, ?> pair3b = Pair.getInstance(1, null);
    
    final Pair<?, ?> pair4a = Pair.getInstance(1, 2);
    final Pair<?, ?> pair4b = Pair.getInstance(1, 2);
    
    assertEquals("Different hashcodes for equal pairs", 
                 pair1a.hashCode(), pair1b.hashCode());
    assertEquals("Different hashcodes for equal pairs", 
                 pair2a.hashCode(), pair2b.hashCode());
    assertEquals("Different hashcodes for equal pairs", 
                 pair3a.hashCode(), pair3b.hashCode());
    assertEquals("Different hashcodes for equal pairs", 
                 pair4a.hashCode(), pair4b.hashCode());
  }
}
