package net.afyre.haha.demo.util;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.Iterator;

/**
 * Helper methods related to comparisons and equality testing.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class CmpUtil {
  /** Singleton instance. */
  private static final CmpUtil instance = new CmpUtil();
  
  /** Constructor. */
  private CmpUtil() { }
  
  /**
   * Get an instance of CmpUtil.
   * @return Instance.
   */
  public static CmpUtil getInstance() {
    return instance;
  }
  
  /**
   * Compare two objects (null-safe).
   * @param <A> Type of the first object.
   * @param <B> Type of the second object.
   * @param a First object.
   * @param b Second object.
   * @return Comparison result compatible 
   *         with specification of 
   *         {@link Comparable#compareTo}.
   */
  public<A extends Comparable<? super B>, B> 
  int compare(final A a, final B b) {
    int result;
    
    if (a == null) {
      if (b == null) {
        result = 0;
      } else {
        result = -1;
      }
    } else {
      if (b == null) {
        result = 1;
      } else {
        result = a.compareTo(b);
      }
    }
    return result;
  }
  
  /**
   * Compare two pairs of objects.
   * @param <A1> Type of the first component of the first pair.
   * @param <A2> Type of the second component of the first pair.
   * @param <B1> Type of the first component of the second pair.
   * @param <B2> Type of the second component of the second pair.
   * @param a1 First component of the first pair.
   * @param a2 Second component of the first pair.
   * @param b1 First component of the second pair.
   * @param b2 Second component of the second pair.
   * @return Comparison result compatible 
   *         with specification of 
   *         {@link Comparable#compareTo}.
   */
  public<A1 extends Comparable<? super B1>, 
         A2 extends Comparable<? super B2>, B1, B2>
  int compare(final A1 a1, final A2 a2, final B1 b1, final B2 b2) {
    int result;
    
    result = compare(a1, b1);
    if (result == 0) {
      result = compare(a2, b2);
    }
    return result;
  }

  /**
   * Compare two triples of objects.
   * @param <A1> Type of the first component of the first triple.
   * @param <A2> Type of the second component of the first triple.
   * @param <A3> Type of the third component of the first triple.
   * @param <B1> Type of the first component of the second triple.
   * @param <B2> Type of the second component of the second triple.
   * @param <B3> Type of the third component of the second triple.
   * @param a1 First component of the first triple.
   * @param a2 Second component of the first triple.
   * @param a3 Third component of the first triple.
   * @param b1 First component of the second triple.
   * @param b2 Second component of the second triple.
   * @param b3 Third component of the second triple.
   * @return Comparison result compatible 
   *         with specification of 
   *         {@link Comparable#compareTo}.
   */
  public<A1 extends Comparable<? super B1>, 
         A2 extends Comparable<? super B2>, 
         A3 extends Comparable<? super B3>, 
         B1, B2, B3>
  int compare(final A1 a1, final A2 a2, final A3 a3, 
              final B1 b1, final B2 b2, final B3 b3) {
    int result;
    
    result = compare(a1, b1);
    if (result == 0) {
      result = compare(a2, b2);
    }
    if (result == 0) {
      result = compare(a3, b3);
    }
    return result;
  }
  
  /**
   * Compare two tuples of four objects.
   * @param <A1> Type of the first component of the first tuple.
   * @param <A2> Type of the second component of the first tuple.
   * @param <A3> Type of the third component of the first tuple.
   * @param <A4> Type of the fourth component of the first tuple.
   * @param <B1> Type of the first component of the second tuple.
   * @param <B2> Type of the second component of the second tuple.
   * @param <B3> Type of the third component of the second tuple.
   * @param <B4> Type of the fourth component of the second tuple.
   * @param a1 First component of the first tuple.
   * @param a2 Second component of the first tuple.
   * @param a3 Third component of the first tuple.
   * @param a4 Fourth component of the first tuple.
   * @param b1 First component of the second tuple.
   * @param b2 Second component of the second tuple.
   * @param b3 Third component of the second tuple.
   * @param b4 Fourth component of the second tuple.
   * @return Comparison result compatible 
   *         with specification of 
   *         {@link Comparable#compareTo}.
   */
  public<A1 extends Comparable<? super B1>, 
         A2 extends Comparable<? super B2>, 
         A3 extends Comparable<? super B3>, 
         A4 extends Comparable<? super B4>, 
         B1, B2, B3, B4>
  int compare(final A1 a1, final A2 a2, final A3 a3, final A4 a4, 
              final B1 b1, final B2 b2, final B3 b3, final B4 b4) {
    int result;
    
    result = compare(a1, b1);
    if (result == 0) {
      result = compare(a2, b2);
    }
    if (result == 0) {
      result = compare(a3, b3);
    }
    if (result == 0) {
      result = compare(a4, b4);
    }
    return result;
  }
  
  /**
   * Compare two arrays lexicographically.
   * @param <L> Type of array elements (first array).
   * @param <R> Type of array elements (second array).
   * @param a First array.
   * @param b Second array.
   * @return Comparison result compatible 
   *         with specification of 
   *         {@link Comparable#compareTo}.
   */
  public<L extends Comparable<? super R>, R> 
  int compare(final L[] a, final R[] b) {
    int i, result;
    
    if (a == null) {
      if (b == null) {
        result = 0;
      } else {
        result = -1;
      }
    } else if (b == null) {
      result = 1;
    } else {
      result = 0;
      i = 0;
      while (result == 0 && i < a.length && i < b.length) {
        result = compare(a[i], b[i]);
        i = i + 1;
      }
      if (result == 0 && i < a.length) {
        result = 1;
      }
      if (result == 0 && i < b.length) {
        result = -1;
      }
    }
    return result;
  }
  
  /**
   * Compare two iterables lexicographically.
   * @param <L> Type of elements returned by the first iterable.
   * @param <R> Type of elements returned by the second iterable.
   * @param a First iterable.
   * @param b Second iterable.
   * @return Comparison result compatible 
   *         with specification of 
   *         {@link Comparable#compareTo}.
   */
  public<L extends Comparable<? super R>, R> 
  int compare(final Iterable<? extends L> a, 
              final Iterable<? extends R> b) {
    int result;
    final Iterator<? extends L> li;
    final Iterator<? extends R> ri;
    
    if (a == null) {
      if (b == null) {
        result = 0;
      } else {
        result = -1;
      }
    } else if (b == null) {
      result = 1;
    } else {
      result = 0;
      li = a.iterator();
      ri = b.iterator();
      while (result == 0 && li.hasNext() && ri.hasNext()) {
        result = compare(li.next(), ri.next());
      }
      if (result == 0 && li.hasNext()) {
        result = 1;
      }
      if (result == 0 && ri.hasNext()) {
        result = -1;
      }
    }
    return result;
  }
  
  /**
   * Null-safe equality test.
   * 
   * Simply call equals() after checking for null references.
   * 
   * @param a First object.
   * @param b Second object.
   * @return True if both objects are null or a.equals(b), 
   *         false otherwise.
   */
  public boolean safeEquals(final Object a, final Object b) {
    final boolean result;
    
    if (a == null) {
      result = (b == null);
    } else if (b == null) {
      result = false;
    } else {
      result = a.equals(b);
    }
    
    return result;
  }

  
  /**
   * Null-safe structural equality test for arrays.
   * 
   * @param <T> Type of array elements.
   * @param a First array.
   * @param b Second array.
   * @return True if both objects are null or a.equals(b), 
   *         false otherwise.
   */
  public<T> boolean safeEquals(final T[] a, final T[] b) {
    boolean result;
    
    if (a == null) {
      result = (b == null);
    } else if (b == null) {
      result = false;
    } else if (a.length != b.length) {
      result = false;
    } else {
      result = true;
      // Lengths are equal
      for (int i = 0; i < a.length && result; i++) {
        result = safeEquals(a[i], b[i]);
      }
    }
    
    return result;
  }
  
  /**
   * Null-safe structural equality test for iterables.
   * 
   * @param a First iterable.
   * @param b Second iterable.
   * @return True if both iterables are null
   *         or return equal sequences of objects.
   */
  public boolean iterEquals(final Iterable<?> a, 
                            final Iterable<?> b) {
    boolean result;
    final Iterator<?> li;
    final Iterator<?> ri;
    
    if (a == null) {
      if (b == null) {
        result = true;
      } else {
        result = false;
      }
    } else if (b == null) {
      result = false;
    } else {
      result = true;
      li = a.iterator();
      ri = b.iterator();
      while (result && li.hasNext() && ri.hasNext()) {
        if (!safeEquals(li.next(), ri.next())) {
          result = false;
        }
      }
      if (li.hasNext() || ri.hasNext()) {
        result = false;
      }
    }
    return result;
  }
  
  /**
   * The method checks if files of the given names differ.
   * 
   * @param src  the first file to compare
   * @param trgt the second file to compare
   * @return true if files differ or exception occurred
   */
  public static boolean differFiles(String src, String trgt) {
	  	try {
			FileInputStream sfile = new FileInputStream(src);
			FileInputStream tfile = new FileInputStream(trgt);
			int b=1;
			while (b>0) {
				b=sfile.read();
				int c=tfile.read();
				if (b!=c) {
					sfile.close();
					tfile.close();
					return true;
				}
			}
			sfile.close();
			tfile.close();
	  	} catch (IOException e) {
			return true;
		}
		return false;
	}
  
}
