package net.afyre.haha.demo.util;

import java.util.Iterator;

/**
 * Iterable which prepends an element to the beginning 
 * of another iterable.
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 * @param <T> Type of elements returned by the iterable.
 */
public final class ConsIterable<T> implements Iterable<T> {
  /** First element. */
  private final T head;
  
  /** Rest of the iterable. */
  private final Iterable<? extends T> tail;
  
  /**
   * Constructor.
   * @param head First element.
   * @param tail Rest of the iterable.
   * @throws NullPointerException If tail is null.
   */
  public ConsIterable(final T head, final Iterable<? extends T> tail) {
    if (tail == null) {
      throw new NullPointerException("tail");
    }
    this.head = head;
    this.tail = tail;
  }
  
  /**
   * Create an instance of ConsIterable.
   * @param <T> Type of elements returned by the iterable.
   * @param head First element.
   * @param tail Rest of the iterable.
   * @throws NullPointerException If tail is null.
   */
  public static<T> ConsIterable<T> getInstance(
      final T head, final Iterable<? extends T> tail) {
    return new ConsIterable<T>(head, tail);
  }
  
  /** {@inheritDoc} */
  public Iterator<T> iterator() {
    return new ConsIterator();
  }
  
  /** Iterator implementation. */
  private final class ConsIterator implements Iterator<T> {
    /** Tail iterator, set to null until head is returned. */
    private Iterator<? extends T> it;
    
    public ConsIterator() {
      this.it = null;
    }
    
    /** {@inheritDoc} */
    @Override
    public boolean hasNext() {
      if (it == null) {
        return true;
      } else {
        return it.hasNext();
      }
    }

    /** {@inheritDoc} */
    @Override
    public T next() {
      if (it == null) {
        it = tail.iterator();
        return head;
      }
      return it.next();
    }

    /** {@inheritDoc} */
    @Override
    public void remove() {
      throw new UnsupportedOperationException();
    }
    
  }
}
