package net.afyre.haha.demo.util;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;

/**
 * An output stream wrapper which forwards all data 
 * to other output streams.
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 */
public final class TeeOutputStream extends OutputStream {
  /** List of output streams. */
  private final OutputStream[] outputs;
  
  /** Should the output streams be closed by this wrapper? */
  private final boolean closeOutputs;
  
  /**
   * Create a wrapper for a given list of output streams.
   * All streams are closed when the wrapper is closed.
   * @param outputs List of output streams.
   * @throws NullPointerException If the list of streams contains a null,
   *                              or is null.
   */
  public static final TeeOutputStream getInstance(
      final OutputStream... outputs) {
    return new TeeOutputStream(true, outputs);
  }  
  
  /**
   * Create a wrapper for a given list of output streams.
   * @param closeOutputs Should the output streams be closed by this wrapper?
   * @param outputs List of output streams.
   * @throws NullPointerException If the list of streams contains a null,
   *                              or is null.
   */
  public static final TeeOutputStream getInstance(
      final boolean closeOutputs,
      final OutputStream... outputs) {
    return new TeeOutputStream(closeOutputs, outputs);
  }
  
  /**
   * Constructor.
   * @param closeOutputs Should the output streams be closed by this wrapper?
   * @param outputs List of output streams.
   * @throws NullPointerException If the list of streams contains a null,
   *                              or is null.
   */
  private TeeOutputStream(final boolean closeOutputs,
                          final OutputStream... outputs) {
    if (outputs == null) {
      throw new NullPointerException("outputs");
    }
    for (int i = 0; i < outputs.length; i++) {
      if (outputs[i] == null) {
        throw new NullPointerException("outputs[" + i +"]");
      }
    }
    this.closeOutputs = closeOutputs;
    this.outputs = Arrays.copyOf(outputs, outputs.length);
  }

  /** @{inheritDoc} */
  @Override
  public void flush() throws IOException {
    for (final OutputStream output : outputs) {
      output.flush();
    }
  }
  
  /** @{inheritDoc} */
  @Override
  public void close() throws IOException {
    if (closeOutputs) {
      for (final OutputStream output : outputs) {
        output.close();
      }
    }
  }
  
  /** @{inheritDoc} */
  @Override
  public void write(final int b) throws IOException {
    for (final OutputStream output : outputs) {
      output.write(b);
    }
  }
  
  /** @{inheritDoc} */
  @Override
  public void write(final byte[] buffer, 
                    int offset, 
                    int length) throws IOException {
    for (final OutputStream output : outputs) {
      output.write(buffer, offset, length);
    }
  }
  

  /** @{inheritDoc} */
  @Override
  public void write(final byte[] buffer) throws IOException {
    for (final OutputStream output : outputs) {
      output.write(buffer);
    }
  }
}
