package net.afyre.haha.demo.vcgen;

import java.util.List;

import net.afyre.haha.demo.haha.Cond;
import net.afyre.haha.demo.haha.Expression;
import net.afyre.haha.demo.haha.Loop;
import net.afyre.haha.demo.haha.LoopInvariant;
import net.afyre.haha.demo.haha.Postcondition;
import net.afyre.haha.demo.haha.Precondition;
import net.afyre.haha.demo.solver.expr.formulae.Formula;
import net.afyre.haha.demo.util.Function;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.xtext.nodemodel.ICompositeNode;
import org.eclipse.xtext.nodemodel.util.NodeModelUtils;

import com.google.common.collect.ImmutableList;

/** 
 * Function used to convert AST expressions to formulae. 
 * 
 * @author Tadeusz Sznuk <tsznuk@mimuw.edu.pl>
 *
 */
public final class ExprToDocumentedFormula 
    implements Function<Expression, DocumentedFormula> {

  /** Object used to convert AST expressions to terms. */
  private final ExpressionUtil exprUtil = ExpressionUtil.getInstance();

  /**
   * Find suitable name for a condition related to 
   * correctness of a given assertion / postcondition.
   * @param assertion The assertion.
   * @return Name.
   */
  public static String getAssertionName(final Expression assertion) {
    String name = getObjectName(assertion.eContainer());
    if (name == null) {
      final String kind = getConditionKind(assertion.eContainer());
      name = kind + "_at_L" + getLine(assertion);
    }
    return name;
  }
  
  /**
   * Find suitable comment for a condition related to 
   * correctness of a given assertion / postcondition.
   * @param assertion The assertion.
   * @return Name.
   */
  private static String getAssertionComment(final Expression assertion) {
    String name = getObjectName(assertion.eContainer());
    if (name == null) {
      name = getConditionKind(assertion.eContainer());
    }
    return name + " at line " + getLine(assertion);
  }
  
  /**
   * Get 'name' attribute of an EObject.
   * Returns 'null' if there is no such attribute, 
   * or if the passed object reference is null.
   * @param obj An object (possibly null).
   * @return Name.
   */
  private static String getObjectName(final EObject obj) {
    if (obj == null) {
      return null;
    }
    final EStructuralFeature feature = 
      obj.eClass().getEStructuralFeature("name");
    if (feature == null) { return null; }
    final Object name = obj.eGet(feature);
    if (name instanceof String) {
      return (String)name;
    } else {
      return null;
    }
  }
  
  /**
   * Get type of a condition: assertion, precondition, invariant, ...
   * @param obj Condition.
   * @return Type (as string
   */
  private static String getConditionKind(final EObject obj) {
    if (obj instanceof Precondition) {
      return "Precondition";
    } else if (obj instanceof Postcondition) {
      return "Postcondition";
    } else if (obj instanceof LoopInvariant) {
      return "Invariant";
    } else if (obj instanceof Loop || obj instanceof Cond) {
      return "Condition";
    } else {
      return "Assertion";
    }
  }

  // TODO: This method is duplicated - move it to a helper class.
  // Get line number from a model object.
  // Return 'X' if line number information is not available.
  private static String getLine(final EObject obj) {
    ICompositeNode node = NodeModelUtils.getNode(obj);
    if (node == null) {
      return "X";
    } else {
      return Integer.toString(node.getStartLine());
    }
  }
  
  /** {@inheritDoc} */
  @Override
  public DocumentedFormula apply(final Expression arg) {
    final Formula formula = exprUtil.translateFormula(arg);
    final String name = getAssertionName(arg);
    final String comment = getAssertionComment(arg);
    final List<EObject> source;
    if (arg.eContainer() != null) {
      source = ImmutableList.of((EObject)arg.eContainer());
    } else {
      source = ImmutableList.of((EObject)arg);
    }
    return new DocumentedFormula(formula, name, source, comment);
  }

}
